#encoding:utf-8

# Copyright (C) 2006 Amit Man <amit.man@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# SentenceSplitter was taken from SentenceSplitter.py
# Copyright (C) 2004  Mickel Grönroos <Mickel.Gronroos@csc.fi>
# licensed under the GNU General Public License version 2 or later.

"""
    This plugin will display a GTK+ dialog with 'did you know..' information
    from Wikipedia when a song starts playing in Exaile
"""

import gtk
import gtk.glade
import pygtk
import xl.plugins as plugins
import HTMLParser
import urllib
import urllib2
import random
import threading
import os, re, locale, string
from gettext import gettext as _
import xl.path

PLUGIN_NAME = _("Did You Know")
PLUGIN_AUTHORS = ['Amit Man <amit.man@gmail.com>']
PLUGIN_VERSION = '0.4.4'
PLUGIN_DESCRIPTION = _(r"""Displays a 'did you know..' information from Wikipedia.
\n\nPress the 'i' icon the show the popup window. Double click on popup window to hide it.""")
PLUGIN_ENABLED = False

b = gtk.Button()
PLUGIN_ICON = b.render_icon('gtk-info', gtk.ICON_SIZE_MENU)
b.destroy()

# For HTTP requests. The APP variable doesn't exist yet, so we set this in
# initialize().
USER_AGENT = None

CONNS = plugins.SignalContainer()

## Default module variables
if os.name == "posix":
    LOCALE = "en_FI"
else:
    LOCALE = "en"
ESCAPE = [(".", "_PERIOD_"), (":", "_COLON_")]

class SentenceSplitter(object):
    """The SentenceSplitter class."""

    def __init__(self,
                 loc=LOCALE,
                 abbreviations=[],
                 escape=ESCAPE):
        """Construct a SentenceSplitter object.

        Parameters:
        1. loc (a string or tuple to feed locale.setlocale()
           (Default: """+str(LOCALE)+""")
        2. abbreviations (a "stop list" of abbreviations that should not be
           split)
           (Default: [])
        3. escape (a sequence of tuples to escape punctuation in the stop list)
           (Default: """+str(ESCAPE)+""")
        """

        ## Set prerequisites
        self.setLocale(loc)
        self.setAbbreviations(abbreviations)
        self.setEscape(escape)

        ## The regular expression matching sentence boundaries
        self._regexpstring   = ("""([\!\"#\'\(\)\.\?]+) ([\"\'\(\-]*\s?["""+
                                string.uppercase+string.digits+"""])""")
        ## A pattern to go with the regular expression for splitting
        ## a chunk of text into sentences
        self._replacepattern = r"\1\n\2"

        ## Compile the regular expression object
        self._regexpobject   = re.compile(self._regexpstring)

    def setLocale(self, loc):
        """Sets the locale. Parameter must be in the format accepted
        by locale.setlocale()."""
        locale.setlocale(locale.LC_ALL)

    def getLocale(self):
        """Returns the current locale."""
        return locale.getlocale()
    
    def setAbbreviations(self, abbreviations):
        """Sets the abbreviation "stop list", i.e. a list of abbreviations
        that should not trigger a split."""
        self._abbreviations = abbreviations

    def getAbbreviations(self):
        """Returns the "stop list" of abbreviations."""
        try:
            return self._abbreviations
        except:
            return []

    def setEscape(self, escape):
        """Sets the the escape handling, i.e. how the punctuation characters
        in the abbreviation stop list should be escaped before splitting
        and turned back to after splitting. The parameter should be a sequence
        of tuples.
        (Example: escape=[(".", "_PERIOD_"), (":", "_COLON_")]
        """
        self._escape = escape

    def getEscape(self):
        """Returns the current sequence of tuples used for escaping
        punctuation in the abbreviations in the stop list."""
        try:
            return self._escape
        except:
            return []

    def split(self, text):
        """Splits a chunk of text into a list of sentences."""

        ## First "escape" all abbreviations in a rather ugly manner
        for abbrev in self.getAbbreviations():
            if text.count(abbrev):
                for t_escapemapping in self.getEscape():
                    escabbrev = abbrev.replace(t_escapemapping[0],
                                               t_escapemapping[1])
                    text = text.replace(abbrev, escabbrev)
                
        ## Then try doing the replace given the regular expression
        ## and the replace pattern
        sentencestring = self._regexpobject.sub(self._replacepattern, text)

        ## Now "unescape" the abbreviations
        for t_escapemapping in self.getEscape():
            if sentencestring.count(t_escapemapping[1]):
                sentencestring = sentencestring.replace(t_escapemapping[1],
                                                         t_escapemapping[0])

        ## Split sentencestring on newlines and return the list
        return sentencestring.split("\n")

class content(HTMLParser.HTMLParser,threading.Thread):
    
    def __init__(self,the_page):
        threading.Thread.__init__(self)
        HTMLParser.HTMLParser.__init__(self)
        self.sentencesplitter=SentenceSplitter()
        self.in_paragraph=False
        self.num_of_para=-1 # when we first meet tag p, we will be 0
        self.paraString=[]
        self.paraSentances=[]
        self.end_of_intro=False
        self.cant_find_page=False
        self.gotimage=False
        
        self.feed(the_page)
        self.parse_to_sentances()
        

    def handle_starttag(self,tag,attrs):
        if tag == 'p':
            #if not self.end_of_intro:
                self.in_paragraph=True
                self.num_of_para=self.num_of_para+1
                self.paraString.append('')
            #print 'len of parastring: %d' % len(self.paraString)
            #print 'start paragraph: %d' % self.num_of_para
       
        if tag =='table':
            if  attrs[0][0] == 'id' and attrs[0][1]=='toc':
                self.end_of_intro=True

        if tag =='img':
            if not self.gotimage:
                for desc in attrs:
                    if desc[0]=='src':
                        src=desc[1]
                        print src
                        if src[-3:] != 'jpg':
                            print 'breaking'
                            break
                        print 'good pic'
                        self.gotimage=True
                        print src
                        
                        
                        headers = { 'User-Agent' : USER_AGENT }
                        req = urllib2.Request(src, None, headers)
                        response = urllib2.urlopen(req)
                        image = response.read()
                        imagefile=file('didyouknow_tmp.img','w')
                        imagefile.write(image)
                        imagefile.close
                    
    def handle_endtag(self,tag):
        if tag =='p':
            self.in_paragraph=False
            #if the last paragraph was empty
            if not len(self.paraString[self.num_of_para]):
                self.num_of_para=self.num_of_para-1
                self.paraString.pop()

    def handle_data(self,data):
        if data=='Wikipedia does not have an article with this exact name.':
            print "cant find wikipedia page"
            self.no_page()
        
        if self.in_paragraph: #and not self.end_of_intro:
            self.paraString[self.num_of_para]=self.paraString[self.num_of_para]+data
        

    def parse_to_sentances(self):
        for s in self.paraString:
            self.paraSentances.append(self.sentencesplitter.split(s))


    def get_random_fact(self):
        if self.cant_find_page:
            return "Cant find info on wikipedia"
        else:
            s=''
            random_para=random.randint(0,len(self.paraSentances)-1)
            random_sen=random.randint(0,len(self.paraSentances[random_para])-1)
            if random_sen > 0:
                s=self.half_sentence(self.paraSentances[random_para][random_sen-1],'begin')+' '
            s=s+self.paraSentances[random_para][random_sen]
            if random_sen < len(self.paraSentances[random_para])-1:
                    s=s+' '+self.half_sentence(self.paraSentances[random_para][random_sen+1],'end')
            return s

    def half_sentence(self,sentence,side='begin'):
        wa=sentence.split(' ')
        word_count=len(wa)
        if side=='end':
            wa=wa[0:(word_count/2)]+['... ']
        if side == 'begin':
            wa=[' ...']+wa[(word_count/2):word_count]
        return (' '.join(wa))


    def no_page(self):
        self.cant_find_page=True








ig=None
newSong=True
songName=''
APP = None
BUTTON = None
MENU_ITEM = None
TIPS = gtk.Tooltips()

D=None
L=None

def initialize():
    """
        Called when the plugin is enabled
    """
    global ig
    global BUTTON
    global USER_AGENT
    

    USER_AGENT = '%s/%s (Exaile %s)' % \
        (plugins.name(__file__), PLUGIN_VERSION, APP.get_version())

    BUTTON = gtk.Button()
    TIPS.set_tip(BUTTON, _("Toggle 'Did you know..' popup"))
    image = gtk.Image()
    image.set_from_stock('gtk-info', gtk.ICON_SIZE_MENU)
    BUTTON.set_image(image)
    BUTTON.set_size_request(32, 32)
    ig=InfoGiver(APP,BUTTON)
    BUTTON.connect('clicked', ig.toggle_popups)

    APP.xml.get_widget('rating_toolbar').pack_start(BUTTON)
    BUTTON.show()

    #CONNS.connect(APP, 'play-track', play_track)
    CONNS.connect(APP.player, 'stop-track', stop_track)
    CONNS.connect(APP, 'track-information-updated',track_information_updated)
    return True

def destroy():
    """
        Called when the plugin is disabled
    """
    global BUTTON

    if BUTTON:
        BUTTON.hide()
        BUTTON.destroy()
        BUTTON = None
    CONNS.disconnect_all()

def play_track(exaile, track):
    """
        Called when playback on a track starts ("play-track" event)
    """
    global ig
    
    if track.type != 'stream':
        ig.get_song_info(exaile,track)

def stop_track(exaile, track):
    global ig
    global songName
    songName=''
    
def track_information_updated(exaile):

    global ig
    global songName
    global enable_popups
    track=exaile.player.current
    if (ig.enable_popups):
        
        #if track.type == 'stream':
        if songName != track.title:
            songName=track.title
            print "TRYING: to find on %s " % track.artist
            ig.get_song_info(exaile,track)
            ig.show_random_info()

class InfoGiver:

    def __init__(self,APP,toggle_button):

        self.toggle_button=toggle_button
        self.c=None
        self.enable_popups=False
        wtree=gtk.glade.xml_new_from_buffer(XML_STRING, len(XML_STRING))
        self.D=wtree.get_widget('window')
        self.L=wtree.get_widget('label')
        self.image=wtree.get_widget('image')
        self.vbox1=wtree.get_widget('vbox1')
        self.title=wtree.get_widget('title')
        self.event=wtree.get_widget('event')
        self.sep_event=wtree.get_widget('sep_event')
        self.title_event=wtree.get_widget('title_event')
        self.event.connect('button_press_event', self.start_dragging)
        self.event.connect('button_press_event', self.close_by_double_click)
        self.event.connect('button_release_event', self.stop_dragging)
        self.sep_event.connect('button_press_event', self.start_dragging)
        self.sep_event.connect('button_press_event', self.close_by_double_click)
        self.title_event.connect('button_press_event', self.close_by_double_click)
        self.sep_event.connect('button_release_event', self.stop_dragging)
        self.title_event.connect('button_press_event', self.start_dragging)
        self.title_event.connect('button_release_event', self.stop_dragging)
        self.__handler = None
        self.B=wtree.get_widget('button')
        self.D.hide();
        self.D.connect('destroy',self.popup_destroyed)
        self.B.connect('clicked',self.activated)
        bg_color = gtk.gdk.color_parse('#567EA2');
        self.D.modify_bg(gtk.STATE_NORMAL, bg_color)
        self.title_event.modify_bg(gtk.STATE_NORMAL, bg_color)
        self.sep_event.modify_bg(gtk.STATE_NORMAL, bg_color)
        self.event.modify_bg(gtk.STATE_NORMAL, bg_color)
        
#------------------------------------------------
# dragging capability. copied from xlmisc.py


    def close_by_double_click(self,widget,event):
        
        if event.type == gtk.gdk._2BUTTON_PRESS:
            bg_color = gtk.gdk.color_parse('white');
            self.toggle_button.modify_bg(gtk.STATE_NORMAL, bg_color)
            self.enable_popups = False
            self.D.hide()
   

    def start_dragging(self, widget, event):
        """
            Called when the user starts dragging the window
        """
        self.__start = event.x, event.y
        self.__handler = self.D.connect('motion_notify_event',
            self.dragging)
        self.__timeout = None


    def stop_dragging(self, widget, event):
        """
            Called when the user stops dragging the mouse
        """
        global POPUP
        if self.__handler: self.D.disconnect(self.__handler)
        self.__handler = None
#        if self.start_timer:
#            self.__timeout = gobject.timeout_add(4000, self.D.hide)
#        settings = self.exaile.settings
#        (w, h) = self..get_size()
#        (x, y) = self.window.get_position()

#        settings['osd/x'] = x
#        settings['osd/y'] = y
#        settings['osd/h'] = h
#        settings['osd/w'] = w
    
#        POPUP = OSDWindow(self.exaile, get_osd_settings(settings))

    def dragging(self, widget, event):
        """
            Called when the user drags the window
        """
        self.D.move(int(event.x_root - self.__start[0]),
            int(event.y_root - self.__start[1]))

#-------------------------------------------------------------        
        
    def toggle_popups(self,b):
        
        if self.enable_popups == False:
            self.enable_popups = True
            bg_color = gtk.gdk.color_parse('red');
            self.toggle_button.modify_bg(gtk.STATE_NORMAL, bg_color)
            if APP.player.current != None:
                self.get_song_info(APP,APP.player.current)
                self.show_random_info()
            #else:
            #    self.label.set_label('no track is playing yet..')
            #    self.dialog.show_all()
        #moving from enable popups to diable them        
        elif self.enable_popups == True:
            bg_color = gtk.gdk.color_parse('white');
            self.toggle_button.modify_bg(gtk.STATE_NORMAL, bg_color)
            self.enable_popups = False
            self.D.hide()


    def activated(self,b):
        self.show_random_info()
        
    def get_song_info(self,exaile,track):
        if not self.enable_popups:
            return
        
        locale = exaile.settings.get_str('wikipedia_locale', 'en')
        url = "http://%s.wikipedia.org/wiki/%s" % (locale, track.artist)
        url = url.replace(" ", "_")
        headers = { 'User-Agent' : USER_AGENT }
        req = urllib2.Request(url, None, headers)
        response = urllib2.urlopen(req)
        the_page = response.read()
        self.c=content(the_page)
        self.c.start()
        #adding image
        if self.c.gotimage:
            self.image.set_from_file('didyouknow_tmp.img')
        else:
            self.image.set_from_file(xl.path.get_data('images', 'nocover.png'))
            
        self.title.set_label(_('Artist: ')+ track.artist+'\n\n'+self.c.paraSentances[0][0]);


    def show_random_info(self):

        self.D.show()
        self.L.show()
        if not self.enable_popups:
            return
                        
        self.L.set_label(_("Did you know...\n\n%s") % (self.c.get_random_fact()))

        #self.dialog.show_all()
        self.vbox1.resize_children()
            
    def kill_window(self):
        try:
            self.D.destroy()
        except:
            pass
        

    

    def popup_destroyed(self,w):
        self.enable_popups=False
        bg_color = gtk.gdk.color_parse('white');
        self.toggle_button.modify_bg(gtk.STATE_NORMAL, bg_color)


XML_STRING = None
def load_data(zip):
    global XML_STRING
    XML_STRING = zip.get_data('gui.glade')



#!/usr/bin/env python

# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

__version__ = '0.2.14'

import sys

if sys.platform == 'linux2':
    # Set process name.  Only works on Linux >= 2.1.57.
    try:
        import ctypes
        libc = ctypes.CDLL('libc.so.6')
        libc.prctl(15, 'exaile\0')
    except:
        pass

import gobject
gobject.threads_init()

# this stuff is done first so that only the modules required to connect to an
# already loaded exaile (if available) are loaded.  This helps with the speed
# of remote commands, like --next, --prev
import xl.dbusinterface
EXAILE_OPTIONS = xl.dbusinterface.get_options()
DBUS_EXIT = xl.dbusinterface.test(EXAILE_OPTIONS)

# find out if they are asking for help or version
HELP = False
for val in sys.argv[:]:
    if val in ('-h', '--help'):
        HELP = True
        sys.argv.remove(val)
    elif val == '--version':
        print "Exaile version:", __version__
        sys.exit(0)

import os.path

import pygtk
pygtk.require('2.0') # Must be before 'import gtk'
import gtk


# Find out the location of exaile's working directory, and insert it to sys.path
basedir = os.path.dirname(os.path.realpath(__file__))
if not os.path.exists(os.path.join(basedir, "exaile.py")):
    cwd = os.getcwd()
    if os.path.exists(os.path.join(cwd, "exaile.py")):
        basedir = cwd
sys.path.insert(0, basedir)

# Evil heuristic to check whether Exaile is installed, by looking for Makefile.
installed = not os.path.exists(os.path.join(basedir, 'Makefile'))

import xl.path

options, args = EXAILE_OPTIONS.parse_args()
if options.settings: 
    xl.path.set_configdir(options.settings)
xl.path.init(basedir, installed)

from xl import logger

# set up gettext for translations
import gettext, locale
import gtk.glade
try:
    locale.setlocale(locale.LC_ALL, '')
except locale.Error:
    logger.log_exception()
gettext.textdomain('exaile')
gtk.glade.textdomain('exaile')
gettext.bindtextdomain('exaile', xl.path.localedir)
gtk.glade.bindtextdomain('exaile', xl.path.localedir)

from xl import common
gtk.window_set_default_icon_from_file(xl.path.get_data('images', 'icon.png'))

from xl.gui import main as exailemain

import urllib
# set the user agent
urllib.URLopener.version = "Exaile/%s (compatible; Python-urllib)" % \
    __version__ 

def check_dirs():
    """
        Makes sure the required directories have been created
    """
    covers = xl.path.get_config("covers")
    if not os.path.isdir(covers):
        os.mkdir(covers)

def init():
    global exaile
    if HELP:
        EXAILE_OPTIONS.print_help()
        sys.exit(0)

    if DBUS_EXIT:
        sys.exit(0)

    running_checks = ('next', 'prev', 'stop', 'play', 'guiquery', 'get_title',
        'get_artist', 'get_album', 'get_length', 'current_position',
        'inc_vol', 'dec_vol', 'get_volume', 'query')

    # check passed arguments for options that require exaile to currently be
    # running
    for check in running_checks:
        if getattr(options, check):
            print "No running Exaile instance found."
            sys.exit(1)

    check_dirs()

    logger.log("Exaile " + __version__)
    exaile = exailemain.ExaileWindow(options, xl.path.firstrun)

if __name__ == "__main__": 
    try:
        init()
        gtk.main()
    except SystemExit:
        raise
    except Exception:
        logger.log_exception()



# vim: ts=4
###
#
# Listen is the legal property of mehdi abaakouk <theli48@gmail.com>
# Copyright (c) 2006 Mehdi Abaakouk
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
###

####
##
## This code taken from Listen media player: http://www.listen-project.org
##
####

import gobject
import gtkmozembed

"""
This class wrap MozEmbed
Use workaround to resolve binding problem:
    - freeze when more than 150ko are render
    - crash when render data and the widget is not realize
"""
class MozClient(gtkmozembed.MozEmbed):
    def __init__(self):
        super(MozClient,self).__init__()
        
        
        self.connect("realize",self.on_realize)
        self.connect("unrealize",self.on_unrealize)
        
        self.__render_id = None
        self.is_realize = False
        
        self.current_data = None
        self.current_url = None
        
        self.dbg = False
        
    def set_data(self,url,data):
        self.print_dbg("set_data",url,len(data))           
        self.current_data = data
        self.current_url = url
        if self.is_realize:
            self.on_realize()
        
    def on_unrealize(self,w=None):
        self.print_dbg("on_unrealize")       
        self.is_realize = False
        
    def on_realize(self,w=None):
        self.print_dbg("on_realize")    
        self.is_realize = True
        if self.current_data and self.current_url:
            if self.__render_id:
                try:gobject.remove_source(self.__render_id)
                except:pass
                self.__render_id = None
                self.close_stream()
            self.open_stream(self.current_url, "text/html")
            
            self.__render_id = gobject.idle_add(self.idle_set_data,self.current_data)
                
    def idle_set_data(self,data):
        self.print_dbg("idle_set_data",len(data))
        buffer = 10*1024
        tmp_data = data[:buffer]
        self.append_data(tmp_data,long(len(tmp_data)))
        tmp_data = data[buffer:]
        if len(tmp_data)>0:
            self.__render_id = gobject.idle_add(self.idle_set_data,tmp_data)
        else:
            self.print_dbg("close_stream")
            self.close_stream()  
            self.__render_id = None
            
    def print_dbg(self,*arg):
        if self.dbg: print "DBG:MozClient:",arg
        
        



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk
from xl import xlmisc, common, library

class StatsDialog(gtk.Dialog):
    """
        Shows various statistics for the Exaile music library
    """
    def __init__(self, parent, exaile):
        """
            Initializes the dialog
        """
        self.set_title(_('Library Statistics'))

        self.main = gtk.VBox()
        self.main.set_border_width(5)

        self.child.add(main)



# Bluetooth plugin for Exaile
# Copyright (C) 2007 Johannes Sasongko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gettext import gettext as _
import subprocess
import gtk
from xl import common, xlmisc

PLUGIN_NAME = _("Send via Bluetooth")
PLUGIN_VERSION = '0.2.1'
PLUGIN_AUTHORS = ["Johannes Sasongko <sasongko@gmail.com>"]
PLUGIN_DESCRIPTION = _(r"""Allows sending files via Bluetooth.\n\nRequires
gnome-obex-send.""")

PLUGIN_ICON = None
PLUGIN_ENABLED = False

def send(files):
    try:
        subprocess.call(['gnome-obex-send'] + files)
    except:
        xlmisc.log_exception()

@common.threaded
def send_threaded(files):
    send(files)

def send_selected(*args):
    tracks = APP.tracks
    if not tracks: return
    selected = tracks.get_selected_tracks() or tracks.songs
    if selected:
        send_threaded([t.io_loc for t in selected if t.type != 'stream'])

MENU_ITEM = None

def initialize():
    global MENU_ITEM
    MENU_ITEM = gtk.MenuItem(_("Send via Bluetooth"))
    MENU_ITEM.connect('activate', send_selected)
    APP.xml.get_widget('tools_menu').get_submenu().append(MENU_ITEM)
    MENU_ITEM.show()
    return True

def destroy():
    global MENU_ITEM
    MENU_ITEM.destroy()
    MENU_ITEM = None



import mutagen.oggvorbis
from xl import xlmisc 

TYPE = 'ogg'
VALID_TAGS = xlmisc.VALID_TAGS

def get_tag(f, tag):
    """
        Gets a specific tag, or if the tag does not exist, it returns an empty
        list
    """
    try:
        return [unicode(t) for t in f[tag]]
    except:
        return [] 

def write_tag(tr):
    try:
        com = mutagen.oggvorbis.OggVorbis(tr.io_loc)
    except mutagen.oggvorbis.OggVorbisHeaderError:
        # FIXME: No use, this fails at the save() call.
        com = mutagen.oggvorbis.OggVorbis()
    com.clear()

    for tag in VALID_TAGS:
        if tr.tags[tag]:
            com[tag] = tr.tags[tag]
            
    com.save(tr.io_loc)

def can_change(tag):
    return tag in VALID_TAGS

def is_multi():
    return True

def fill_tag_from_path(tr):
    """
        Fills the passed in media.Track with tags from the file
    """
    try:
        f = mutagen.oggvorbis.OggVorbis(tr.io_loc)
    except mutagen.oggvorbis.OggVorbisHeaderError:
        return

    tr.length = int(f.info.length)
    
    tr.bitrate = (f.info.bitrate // 33554431) * 1000

    for tag in VALID_TAGS:
        tr.tags[tag] = get_tag(f, tag)



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


import glob, md5, os, re, threading, time, urllib, urllib2
from gettext import gettext as _
import gobject, gtk
import xlmisc, library, common
import xl.path
from lib import ecs

COVER_WIDTH = 100
NOCOVER_IMAGE = xl.path.get_data("images", "nocover.png")

#LOCALES = ['ca', 'de', 'fr', 'jp', 'uk', 'us']

KEY = "15VDQG80MCS2K1W2VRR2" # Adam Olsen's key (synic)
ecs.setLicenseKey(KEY)

"""
    Fetches album covers from Amazon.com
"""

class Cover(dict):
    """
        Represents a single album cover
    """
    def save(self, savepath='.'):
        """
            Saves the image to a file
        """
        if not os.path.isdir(savepath):
            os.mkdir(savepath)

        savepath = "%s%s%s.jpg" % (savepath, os.sep, self['md5'])
        handle = open(savepath, "wb")
        handle.write(self['data'])
        handle.close()
        self['filename'] = savepath

    def filename(self):
        return "%s.jpg" % self['md5']

class CoverFetcherThread(threading.Thread):
    """
        Fetches all covers for a search string
    """
    def __init__(self, search_string, _done_func, fetch_all=False, locale='us'): 
        """
            Constructor expects a search string and a function to call
            when it's _done
        """

        xlmisc.log("new thread created with %s" % search_string)
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self._done = False
        self._done_func = _done_func
        self.search_string = search_string
        self.locale = locale
        self.fetch_all = fetch_all
        ecs.setLocale(locale)

    def abort(self):
        """
            Aborts the download thread. Note that this does not happen
            immediately, but happens when the thread is done blocking on its
            current operation
        """
        self._done = True
    

    def run(self):
        """
            Actually connects and fetches the covers
        """
        start = time.time()
        xlmisc.log("Amazon: cover thread started")

        if self._done: return

        covers = []
        try:
            albums = ecs.ItemSearch(Keywords=self.search_string, SearchIndex="Music", 
                ResponseGroup="ItemAttributes,Images")
        except ecs.NoExactMatches:
            xlmisc.log("No cover found")
            gobject.idle_add(self._done_func, covers)
            return
        except Exception:
            xlmisc.log_exception()
            gobject.idle_add(self._done_func, covers)
            return

        if self._done: return

        # using a for x in range loop here because the ecs.py library
        # sometimes gets an exception when looping through all the items
        for i in range(0, len(albums)):
            try:
                album = albums[i]
            except IndexError: continue

            if self._done: return

            cover = Cover()
            # m.group(1) == hostname
            # m.group(2) == filename
            try:
                url = album.LargeImage.URL
            except AttributeError: continue
            try:
                response = urllib2.urlopen(url)
                
                cover['status'] = 200

                data = response.read()
                if self._done: return
                cover['data'] = data
                cover['md5'] = md5.new(data).hexdigest()

                # find out if the cover is valid
                if len(data) > 1200:
                    covers.append(cover)
                    if not self.fetch_all: break
            except urllib2.HTTPError,e:
                cover['status'] = e.code
            except urllib2.URLError:
                pass

        if len(covers) == 0:
            xlmisc.log("Amazon: Thread done.... *shrug*, no covers found")

        if self._done: return

        end = time.time()
        if end - start < 1:
            difference = 1 - (end - start) + .01
            xlmisc.log("Amazon: sleeping for %s seconds..." % difference)
        else:
            difference = .01

        time.sleep(difference)
        # call after the current pending event in the gtk gui
        gobject.idle_add(self._done_func, covers)

class CoverEventBox(gtk.EventBox):
    def __init__(self, exaile, cover):
        gtk.EventBox.__init__(self)
        self.exaile = exaile
        self.db = exaile.db
        self.cover_menu = CoverMenu(self.exaile, is_current_song=True)
        self.connect('button_press_event', self.cover_clicked)
        self.add(cover)
        self.cover = cover

    def cover_clicked(self, widget, event):
        """
            Called when the cover is clicked on
        """
        track = self.exaile.player.current
        if not track:
            return

        if event.type == gtk.gdk._2BUTTON_PRESS:
            if 'nocover' in self.cover.loc:
                CoverFrame(self.exaile, track)
            else:
                CoverWindow(self.exaile.window, self.cover.loc,
                    _("%(album)s by %(artist)s") %
                    {
                        'album': track.album,
                        'artist': track.artist
                    })
        elif event.button == 3:
            self.cover_menu.menu.popup(None, None, None,
                event.button, event.time)

FETCHER = None
def get_cover_fetcher(exaile):
    """
        gets the cover fetcher instance
    """
    global FETCHER

    try:
        if not CoverFetcher.stopped:
            return FETCHER
    except:
        xlmisc.log_exception()
        pass

    FETCHER = CoverFetcher(exaile)
    return FETCHER

class CoverManager(object):
    """
        Manages album art for Exaile
    """
    def __init__(self, exaile):
        self.cover_thread = None
        self.exaile = exaile
        self.db = exaile.db

    def fetch_covers(self, event):
        """
            Fetches all covers
        """
        fetcher = get_cover_fetcher(self.exaile)
        fetcher.dialog.show_all()

    def got_stream_cover(self,covers):
        xlmisc.log("got stream cover")
        self.exaile.status.set_first(None)
        if len(covers) == 0:
            self.exaile.status.set_first(_("No covers found."), 2000)
        
        for cover in covers:
            if(cover['status'] == 200):
                savepath = xl.path.get_config('covers', 'streamCover.jpg')
                handle = open(savepath, "wb")
                handle.write(cover['data'])
                handle.close()
                self.exaile.cover.set_image(savepath)
                break

    def got_covers(self, covers): 
        """
            Gets called when all covers have been downloaded from amazon
        """
        track = self.exaile.player.current
        artist_id = library.get_column_id(self.db, 'artists', 'name', track.artist)
        album_id = library.get_album_id(self.db, artist_id, track.album)

        self.exaile.status.set_first(None)
        if len(covers) == 0:
            self.exaile.status.set_first(_("No covers found."), 2000)
            self.db.execute("UPDATE albums SET image='nocover' WHERE id=?",
                (album_id,))

        # loop through all of the covers that have been found
        for cover in covers:
            if(cover['status'] == 200):
                cover.save(xl.path.get_config('covers'))
                xlmisc.log(cover['filename'])
                self.exaile.cover.set_image(cover['filename'])
                self.exaile.player.current._tmp_cover = cover['filename']

                self.db.execute("UPDATE albums SET image=?, amazon_image=1 WHERE id=?",
                    (cover['md5'] + ".jpg", album_id))
                
                break

    def check_image_age(self, album_id, image):
        """
            This checks to see if the image is too old for Amazon's ULA, and
            if it is, it refetches the image
        """
        info = os.stat(xl.path.get_config('covers', image))

        max_time = 30 * 24 * 60 * 60 # 1 month
        if time.time() - info[9] > max_time:
            self.exaile.status.set_first(_('Current amazon image is too old, '
                'fetching  a new one'), 2000)
            self.db.execute('UPDATE albums SET image=NULL, amazon_image=0 '
                'WHERE id=?', (album_id,))
            self.fetch_cover(self.exaile.player.current)
            return False
        return True
   
    def fetch_cover(self, track, popup=None): 
        """
            Fetches the cover from the database.  If it can't be found
            there it fetches it from amazon
        """
        w = COVER_WIDTH
        if not popup:
            self.exaile.cover.set_image(NOCOVER_IMAGE)

        # for popups, cover might have already been located, so use it
        if popup and hasattr(track, '_tmp_cover'):
            if track._tmp_cover: return track._tmp_cover

        if track == None: return
        artist_id = library.get_column_id(self.db, 'artists', 'name', track.artist)
        album_id = library.get_album_id(self.db, artist_id, track.album)

        # check to see if a cover already exists
        row = self.db.read_one("albums", "image, amazon_image", 'id=?', (album_id,))

        if row != None and row[0] != "" and row[0] != None:
            if row[0] == "nocover": 
                cover = self.fetch_from_fs(track)
                if cover:
                    if popup: return cover
                    self.exaile.cover.set_image(cover)
                    return
                return NOCOVER_IMAGE

            coverfile = xl.path.get_config('covers', row[0])

            if os.path.isfile(coverfile):

                if popup: return coverfile

                # check to see if we need to recache this image
                if row[1]:
                    if not self.check_image_age(album_id, row[0]): return

                if gtk.gdk.pixbuf_get_file_info(coverfile):
                    track._tmp_cover = coverfile
                    self.exaile.cover.set_image(coverfile)
                else:
                    self.exaile.cover.set_image(NOCOVER_IMAGE)
                return

        cover = self.fetch_from_fs(track)
        if cover:
            track._tmp_cover = cover
            if popup: return cover
            else: self.exaile.cover.set_image(cover)
            return

        if popup != None: return NOCOVER_IMAGE
        self.stop_cover_thread()

        if self.exaile.settings.get_boolean("fetch_covers", True):
            locale = self.exaile.settings.get_str('amazon_locale', 'us')
            if track.type == 'stream':
                xlmisc.log("we got a stream type cover fetch")
                self.cover_thread = CoverFetcherThread("%s %s"\
                    % (track.artist,track.title),
                    self.got_stream_cover, locale=locale)

            else:    
                album = track.album
                if not album:
                    album = track.title
                self.cover_thread = CoverFetcherThread("%s - %s" \
                    % (track.artist, album),
                    self.got_covers, locale=locale)

            self.exaile.status.set_first(_("Fetching cover art from Amazon..."))
            self.cover_thread.start()
            
    def fetch_from_fs(self, track, event=None):
        """
            Fetches the cover from the filesystem (if there is one)
        """
        dir = os.path.dirname(track.loc)

        names = self.exaile.settings.get_list('art_filenames', 
            [u'cover.jpg', u'folder.jpg', u'.folder.jpg', u'album.jpg', u'art.jpg'])
        if not names: return None

        for f in names:
            for f in glob.glob(os.path.join(dir, f.strip())):
                if os.path.isfile(f):
                    return f

        return None

    def stop_cover_thread(self): 
        """
            Aborts the cover thread
        """

        if self.cover_thread != None:
            xlmisc.log("Aborted cover thread")
            self.cover_thread.abort()
            self.cover_thread = None

class CoverWrapper(object):
    """
        Wraps a cover object
    """
    def __init__(self, artist, album, location):
        self.artist = artist
        self.album = album
        self.location = location

    def __str__(self):
        title = "%s" % (self.album)
        if len(title) > 12:
            title = title[0:10] + "..."
        return title

class CoverFetcher(object):
    """
        Fetches all covers in the library
    """
    stopped = True

    def __init__(self, parent):
        """ 
            Initializes the dialog
        """
        self.exaile = parent
        self.db = self.exaile.db
        xml = gtk.glade.XML('exaile.glade', 'CoverFetcher', 'exaile')
        self.icons = xml.get_widget('cover_icon_view')

        self.model = gtk.ListStore(str, gtk.gdk.Pixbuf, object)
        self.icons.set_model(self.model)
        self.icons.set_item_width(90)
        self.artists = None
        self.go = False
        self.icons.set_text_column(0)
        self.icons.set_pixbuf_column(1)
        self.icons.connect('button-press-event',
            self.cover_clicked)
        # Leave blank since it just needs the reference for now
        self.current_path = None
        self.cover_menu = CoverMenu(self.exaile, self.model, self.current_path)
        self.status = xml.get_widget('cover_status_bar')
        self.icons.connect('motion-notify-event',
            self.mouse_move)
        self.icons.connect('leave-notify-event',
            lambda *e: self.status.set_label(''))
        self.dialog = xml.get_widget('CoverFetcher')
        self.dialog.connect('delete-event', self.cancel)

        self.progress = xml.get_widget('fetcher_progress')
        self.label = xml.get_widget('fetcher_label')

        xml.get_widget('fetcher_cancel_button').connect('clicked',
            self.cancel)
        self.stopstart = xml.get_widget('fetcher_stop_button')
        self.stopstart.connect('clicked', self.toggle_running)
        xml.get_widget('fetcher_refresh_button').connect('clicked',
            self.refresh)
        self.current = 0
        self.dialog.show_all()
        # Give it a parent now that it exists.
        self.cover_menu.set_parent(self.dialog)
        xlmisc.finish()
        self.total = self.calculate_total()
        self.label.set_label(_("%d covers left to collect.") % self.total)
        if self.go:
            self.toggle_running(None)

    def refresh(self, event):
        self.model = gtk.ListStore(str, gtk.gdk.Pixbuf, object)
        self.icons.set_model(self.model)
        self.icons.set_item_width(90)
        self.total = self.calculate_total()

    def cancel(self, *event):
        """
            Closes the dialog
        """
        CoverFetcher.stopped = True
        self.dialog.destroy()

    def toggle_running(self, event):
        """
            Toggles the running state of the fetcher
        """
        if CoverFetcher.stopped:
            if not self.artists:
                self.go = True
                self.stopstart.set_use_stock(True)
                self.stopstart.set_label('gtk-stop')
                return
                
            CoverFetcher.stopped = False
            self.stopstart.set_use_stock(True)
            self.stopstart.set_label('gtk-stop')
            self.fetch_next()
        else:
            self.stopstart.set_use_stock(False)
            # TRANSLATORS: Start fetching cover arts
            self.stopstart.set_label(_('Start'))
            self.stopstart.set_image(gtk.image_new_from_stock('gtk-yes',
                gtk.ICON_SIZE_BUTTON))
            CoverFetcher.stopped = True
            if self.cover_thread:
                self.cover_thread.abort()

    def fetch_next(self, event=None):
        """
            Fetches the next cover in line
        """
        if not self.artists:
            self.label.set_label(_("All Covers have been Fetched."))
            self.stopstart.set_sensitive(False)
            return
        self.artist = self.artists[0]
        if CoverFetcher.stopped: return
        
        try:
            self.album = self.needs[self.artist].pop()
        except IndexError:
            del self.needs[self.artist]
            self.artists.remove(self.artist)
            self.fetch_next()
            return

        if not self.needs[self.artist]:
            del self.needs[self.artist]
            self.artists.remove(self.artist)

        locale = self.exaile.settings.get_str('amazon_locale', 'us')
        self.cover_thread = CoverFetcherThread("%s - %s" %
            (self.artist, self.album),
            self.got_covers, locale=locale)
        self.cover_thread.start()
        # TRANSLATORS: Album cover fetching status
        self.label.set_label(_("%(remaining)d left: %(album)s by %(artist)s") % 
            {
                'remaining': (self.total - self.current),
                'album': self.album,
                'artist': self.artist
            })

    def got_covers(self, covers):
        """
            Called when the fetcher thread has gotten all the covers for this
            album
        """
        if self.stopped: return
        artist_id = library.get_column_id(self.db, 'artists', 'name', self.artist)
        album_id = library.get_album_id(self.db, artist_id, self.album)
        if len(covers) == 0:
            self.db.execute("UPDATE albums SET image=? WHERE id=?",
                ('nocover', album_id,))
            
        # loop through all of the covers that have been found
        for cover in covers:
            if(cover['status'] == 200):
                cover.save(xl.path.get_config('covers'))
                xlmisc.log(cover['filename'])

                try:
                    self.db.execute("UPDATE albums SET image=? WHERE id=?", 
                        (cover['md5'] + ".jpg", album_id))
                except:
                    xlmisc.log_exception()

                image = xl.path.get_config('covers', cover['md5'] + ".jpg")
                loc = image
                image = gtk.gdk.pixbuf_new_from_file(image)
                image = image.scale_simple(80, 80, 
                    gtk.gdk.INTERP_BILINEAR)
                
                if self.found.has_key("%s - %s" % (self.artist.lower(), self.album.lower())):
                    iter = self.found["%s - %s" % (self.artist.lower(), self.album.lower())] 
                    object = self.model.get_value(iter, 2)
                    object.location = loc
                    self.model.set_value(iter, 1, image)
                break

        if self.stopped: return
        self.current = self.current + 1
        self.progress.set_fraction(float(self.current) / float(self.total))
        self.fetch_next()

    def mouse_move(self, widget, event):
        """
            Called when the mouse moves in the icon view
        """
        x, y = event.get_coords()
        x = int(x)
        y = int(y)

        path = self.icons.get_path_at_pos(x, y)
        #saves path so that it can be used by the menu to locate the album
        self.current_path = path
        self.cover_menu.current_path = path
        if not path:
            self.status.set_label('')
            return

        iter = self.model.get_iter(path)
        object = self.model.get_value(iter, 2)

        self.status.set_markup("<b>" + common.escape_xml(_("%(album)s by %(artist)s") %
            {
                'album': object.album,
                'artist': object.artist
            }) + "</b>")

    def calculate_total(self):
        """
            Finds the albums that need a cover
        """
        all = self.db.select("SELECT artists.name, albums.name, albums.image "
            "FROM tracks,albums,artists WHERE blacklisted=0 AND type=0 AND ( "
            "artists.id=tracks.artist AND albums.id=tracks.album)"
            " ORDER BY albums.name, artists.name")
        self.needs = dict()

        self.found = dict()
        count = 0
        for (artist, album, image) in all:
            if artist == '<Unknown>' or album == '<Unknown>': continue
            if not self.needs.has_key(artist):
                self.needs[artist] = []
            if album in self.needs[artist]: continue

            if image:
                if "nocover" in image:
                    image = NOCOVER_IMAGE
                else:
                    image = xl.path.get_config('covers', image)
            else:
                self.needs[artist].append(album)
                image = NOCOVER_IMAGE


            title = CoverWrapper(artist, album, image)
            try:
                image = gtk.gdk.pixbuf_new_from_file(image)
            except gobject.GError:
                image = gtk.gdk.pixbuf_new_from_file(NOCOVER_IMAGE)
            image = image.scale_simple(80, 80, 
                gtk.gdk.INTERP_BILINEAR)

            if self.found.has_key("%s - %s" % (artist.lower(), album.lower())):
                continue

            self.found["%s - %s" % (artist.lower(), album.lower())] = \
                self.model.append([title, image, title])
            if count >= 200: 
                count = 0
                t = 0
                for k, v in self.needs.iteritems():
                    t += len(v)
                self.label.set_label(_("%d covers left to collect.") % t)
                xlmisc.finish()
            count += 1

        count = 0
        for k, v in self.needs.iteritems():
            count += len(v)

        self.artists = self.needs.keys()
        self.artists.sort()

        return count

    def cover_clicked(self, widget, event):
        """
            Called when the cover is clicked on
        """
        track = self.current_path
        if not track:
            return
        # On double click (removed item_activated signal)
        if event.type == gtk.gdk._2BUTTON_PRESS:
            iter = self.model.get_iter(self.current_path)
            object = self.model.get_value(iter, 2)
            CoverWindow(self.dialog, object.location, _("%(album)s by %(artist)s") %
            {
                'album': object.album,
                'artist': object.artist
            })
        # Creates the nice Right click menu
        if event.button == 3:
            self.cover_menu.menu.popup(None, None, None,
                event.button, event.time)

class CoverWindow(object):
    """Shows the cover in a simple image viewer"""

    def __init__(self, parent, cover, title=''):
        """Initializes and shows the cover"""
        self.widgets = gtk.glade.XML('exaile.glade', 'CoverWindow', 'exaile')
        self.widgets.signal_autoconnect(self)
        self.cover_window = self.widgets.get_widget('CoverWindow')
        self.layout = self.widgets.get_widget('layout')
        self.toolbar = self.widgets.get_widget('toolbar')
        self.zoom_in = self.widgets.get_widget('zoom_in')
        self.zoom_out = self.widgets.get_widget('zoom_out')
        self.image = self.widgets.get_widget('image')
        self.statusbar = self.widgets.get_widget('statusbar')
        self.scrolledwindow = self.widgets.get_widget('scrolledwindow')
        self.scrolledwindow.set_hadjustment(self.layout.get_hadjustment())
        self.scrolledwindow.set_vadjustment(self.layout.get_vadjustment())
        self.cover_window.set_title(title)
        self.cover_window.set_transient_for(parent)
        self.cover_window_width = 500
        self.cover_window_height = 500 + self.toolbar.size_request()[1] + \
                                   self.statusbar.size_request()[1]
        self.scrolledwindow
        self.cover_window.set_default_size(self.cover_window_width, \
                                           self.cover_window_height)
        self.image_original_pixbuf = gtk.gdk.pixbuf_new_from_file(cover)
        self.image_pixbuf = self.image_original_pixbuf
        self.min_percent = 1
        self.max_percent = 500
        self.ratio = 1.5
        self.image_interp = gtk.gdk.INTERP_BILINEAR
        self.image_fitted = True
        self.set_ratio_to_fit()
        self.update_widgets()
        self.cover_window.show_all()
 
    def available_image_width(self):
        """Returns the available horizontal space for the image"""
        return self.cover_window.get_size()[0]

    def available_image_height(self):
        """Returns the available vertical space for the image"""
        return self.cover_window.get_size()[1] - \
               self.toolbar.size_request()[1] - \
               self.statusbar.size_request()[1]

    def center_image(self):
        """Centers the image in the layout"""
        new_x = max(0, int((self.available_image_width() - \
                            self.image_pixbuf.get_width()) / 2))
        new_y = max(0, int((self.available_image_height() - \
                            self.image_pixbuf.get_height()) / 2))
        self.layout.move(self.image, new_x, new_y)

    def update_widgets(self):
        """Updates image, layout, scrolled window, tool bar and status bar"""
        if self.cover_window.window:
            self.cover_window.window.freeze_updates()
        self.apply_zoom()
        self.layout.set_size(self.image_pixbuf.get_width(), \
                             self.image_pixbuf.get_height())
        if self.image_fitted or \
           (self.image_pixbuf.get_width() == self.available_image_width() and \
           self.image_pixbuf.get_height() == self.available_image_height()):
            self.scrolledwindow.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
        else:
            self.scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC,
                                           gtk.POLICY_AUTOMATIC)
        percent = int(100 * self.image_ratio)
        message = str(self.image_original_pixbuf.get_width()) + " x " + \
                      str(self.image_original_pixbuf.get_height()) + \
                      " pixels " + str(percent) + '%'
        self.zoom_in.set_sensitive(percent < self.max_percent)
        self.zoom_out.set_sensitive(percent > self.min_percent)
        self.statusbar.pop(self.statusbar.get_context_id(''))
        self.statusbar.push(self.statusbar.get_context_id(''), message)
        self.image.set_from_pixbuf(self.image_pixbuf)
        self.center_image()
        if self.cover_window.window:
            self.cover_window.window.thaw_updates()

    def apply_zoom(self):
        """Scales the image if needed"""
        new_width = int(self.image_original_pixbuf.get_width() * \
                        self.image_ratio)
        new_height = int(self.image_original_pixbuf.get_height() * \
                         self.image_ratio)
        if new_width != self.image_pixbuf.get_width() or \
           new_height != self.image_pixbuf.get_height(): 
            self.image_pixbuf = self.image_original_pixbuf.scale_simple(new_width, \
                                  new_height, self.image_interp)

    def set_ratio_to_fit(self):
        """Calculates and sets the needed ratio to show the full image"""
        width_ratio = float(self.image_original_pixbuf.get_width()) / \
                            self.available_image_width()
        height_ratio = float(self.image_original_pixbuf.get_height()) / \
                             self.available_image_height()
        self.image_ratio = 1 / max(1, width_ratio, height_ratio)

    def cover_window_destroy(self, widget):
        self.cover_window.destroy()

    def zoom_in_clicked(self, widget):
        self.image_fitted = False
        self.image_ratio *= self.ratio 
        self.update_widgets()

    def zoom_out_clicked(self, widget):
        self.image_fitted = False
        self.image_ratio *= 1 / self.ratio
        self.update_widgets()

    def zoom_100_clicked(self, widget):
        self.image_fitted = False
        self.image_ratio = 1
        self.update_widgets()

    def zoom_fit_clicked(self, widget):
        self.image_fitted = True
        self.set_ratio_to_fit()
        self.update_widgets()

    def cover_window_size_allocate(self, widget, allocation):
        if self.cover_window_width != allocation.width or \
           self.cover_window_height != allocation.height:
            if self.image_fitted:
                self.set_ratio_to_fit()
            self.update_widgets()
            self.cover_window_width = allocation.width
            self.cover_window_height = allocation.height

class CoverFrame(object):
    """
        Fetches all album covers for a string, and allows the user to choose
        one out of the list
    """
    def __init__(self, parent, track, search=False, exaile_parent=None):
        """
            Expects the parent control, a track, an an optional search string
        """
        self.xml = gtk.glade.XML('exaile.glade', 'CoverFrame', 'exaile')
        self.window = self.xml.get_widget('CoverFrame')
        self.window.set_title("%s - %s" % (track.artist, track.album))

        # This basically has to be done for exaile to pass the database
        # correctly also for exaile to not be the parent of the searcher

        if (exaile_parent == None):
            self.window.set_transient_for(parent.window)
            self.parent = parent.window
            self.exaile = parent
            self.db = self.exaile.db
        else:
            self.window.set_transient_for(parent)
            self.parent = parent
            self.exaile = exaile_parent

        self.track = track
        self.db = self.exaile.db
        self.prev = self.xml.get_widget('cover_back_button')
        self.prev.connect('clicked', self.go_prev)
        self.prev.set_sensitive(False)
        self.next = self.xml.get_widget('cover_forward_button')
        self.next.connect('clicked', self.go_next)
        self.xml.get_widget('cover_newsearch_button').connect('clicked',
            self.new_search)
        self.xml.get_widget('cover_cancel_button').connect('clicked',
            lambda *e: self.window.destroy())
        self.ok = self.xml.get_widget('cover_ok_button')
        self.ok.connect('clicked',
            self.on_ok)
        self.box = self.xml.get_widget('cover_image_box')
        self.cover = xlmisc.ImageWidget()
        self.cover.set_image_size(350, 350)
        self.box.pack_start(self.cover, True, True)

        self.last_search = "%s - %s" % (track.artist, track.album)

        if not search:
            locale = self.exaile.settings.get_str('amazon_locale', 'us')
            CoverFetcherThread("%s - %s" % (track.artist, track.album),
                self.covers_fetched, locale=locale).start()
        else:
            self.new_search()

    def new_search(self, widget=None):
        """
            Creates a new search string
        """
        dialog = common.TextEntryDialog(self.parent,
            _("Enter the search text"), _("Enter the search text"))
        dialog.set_value(self.last_search)
        result = dialog.run()
        if result == gtk.RESPONSE_OK:
            self.last_search = dialog.get_value()
            self.exaile.status.set_first(
                _("Searching for %s...") % self.last_search)
            self.window.hide()

            locale = self.exaile.settings.get_str('amazon_locale', 'us')
            CoverFetcherThread(self.last_search,
                self.covers_fetched, True, locale=locale).start()

    def on_ok(self, widget=None):
        """
            Chooses the current cover and saves it to the database
        """
        track = self.track
        cover = self.covers[self.current]
        artist_id = library.get_column_id(self.db, 'artists', 'name', track.artist)
        album_id = library.get_album_id(self.db, artist_id, track.album)

        self.db.execute("UPDATE albums SET image=? WHERE id=?", (cover.filename(),
            album_id))

        if track == self.exaile.player.current:
            self.exaile.cover_manager.stop_cover_thread()
            self.exaile.cover.set_image(
                xl.path.get_config('covers', cover.filename()))
        self.window.destroy()

    def go_next(self, widget):
        """
            Shows the next cover
        """
        if self.current + 1 >= len(self.covers): return
        self.current = self.current + 1
        self.show_cover(self.covers[self.current])
        if self.current + 1 >= len(self.covers):
            self.next.set_sensitive(False)
        if self.current - 1 >= 0:
            self.prev.set_sensitive(True)

    def go_prev(self, widget):
        """
            Shows the previous cover
        """
        if self.current - 1 < 0: return
        self.current = self.current - 1
        self.show_cover(self.covers[self.current])

        if self.current + 1 < len(self.covers):
            self.next.set_sensitive(True)
        if self.current - 1 < 0:
            self.prev.set_sensitive(False)

    def show_cover(self, c):
        """
            Shows the current cover
        """
        c.save(xl.path.get_config('covers') + os.sep)

        xlmisc.log(c.filename())

        self.cover.set_image(xl.path.get_config('covers', c.filename()))

        self.window.show_all()

    def covers_fetched(self, covers):
        """
            Called when the cover fetcher thread has fetched all covers
        """
        self.exaile.status.set_first(None)
        self.covers = covers
        self.prev.set_sensitive(False)

        if not covers:
            common.error(self.parent, _("Sorry, no covers were found."))
            self.next.set_sensitive(False)
            self.ok.set_sensitive(False)
            self.window.show_all()
            return

        self.next.set_sensitive(len(covers) > 1)
        self.ok.set_sensitive(True)

        self.current = 0

        self.show_cover(self.covers[self.current])

class CoverMenu(xlmisc.Menu):

    def __init__(self, exaile, model=None, current_path=None, is_current_song=False):
        self.exaile = exaile
        self.db = self.exaile.db
        self.model = model
        self.current_path = current_path
        self.is_current = is_current_song
        self.menu = xlmisc.Menu()
        self.cover_full = self.menu.append(_("View Full Image"),
            self.cover_menu_activate)
        self.cover_fetch = self.menu.append(_("Fetch from Amazon"),
            self.cover_menu_activate)
        self.cover_search = self.menu.append(_("Search Amazon"),
            self.cover_menu_activate)
        self.cover_custom = self.menu.append(_("Set Custom Image"),
            self.cover_menu_activate)
        self.remove_cover_menu = self.menu.append(_("Remove Cover"),
            self.cover_menu_activate)

    def cover_menu_activate(self, item, user_param=None):
        """
            Called when one of the menu items in the album cover popup is
            selected
        """

        if self.is_current:
            self.mother = self.exaile.window
            track = self.exaile.player.current
            loc = self.exaile.player.current
        else:
            iter = self.model.get_iter(self.current_path)
            track = self.model.get_value(iter, 2)
            loc = track.location

        if item == self.cover_fetch:
            self.exaile.status.set_first(_("Fetching from Amazon..."))
            CoverFrame(self.mother, track, exaile_parent=self.exaile)
        elif item == self.cover_search:
            CoverFrame(self.mother, track, True, exaile_parent=self.exaile)
        elif item == "showcover" or item == self.cover_full:
            if self.is_current:
                loc = self.exaile.cover.loc

            if "nocover" in loc: return
            CoverWindow(self.mother, loc, _("%(album)s by %(artist)s") %
            {
                'album': track.album,
                'artist': track.artist
            })
        elif item == self.cover_custom:

            dialog = gtk.FileChooserDialog(_("Choose an image"), self.mother,
                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                gtk.STOCK_OPEN, gtk.RESPONSE_OK))
            dialog.set_current_folder(self.exaile.get_last_dir())

            # Image Files
            filter = gtk.FileFilter()
            filter.set_name(_("Image Files"))
            for pattern in ('*.jpg', '*.jpeg', '*.gif', '*.png'):
                filter.add_pattern(pattern)
            dialog.add_filter(filter)
            # All Files
            filter = gtk.FileFilter()
            filter.set_name(_("All Files"))
            filter.add_pattern('*')
            dialog.add_filter(filter)

            result = dialog.run()
            dialog.hide()

            if result == gtk.RESPONSE_OK:
                self.exaile.last_open_dir = dialog.get_current_folder()
                handle = open(dialog.get_filename(), "rb")
                data = handle.read()
                handle.close()

                (f, ext) = os.path.splitext(dialog.get_filename())
                newname = md5.new(data).hexdigest() + ext
                handle = open(xl.path.get_config('covers', newname), 'wb')
                handle.write(data)
                handle.close()

                path_id = library.get_column_id(self.db, 'paths', 'name',
                    loc)
                artist_id = library.get_column_id(self.db, 'artists', 'name',
                    track.artist)
                album_id  = library.get_album_id(self.db, artist_id,
                    track.album)

                xlmisc.log(newname)

                self.db.execute("UPDATE albums SET image=? WHERE id=?",
                    (newname, album_id))

                if track == self.exaile.player.current:
                    self.exaile.cover_manager.stop_cover_thread()
                    self.exaile.cover.set_image(
                        xl.path.get_config('covers', newname))
        elif item == self.remove_cover_menu:
            artist_id = library.get_column_id(self.db, 'artists', 'name', track.artist)
            album_id = library.get_album_id(self.db, artist_id, track.album)
            #print album_id  #testing to check that right album is removed

            # Sets image to NULL so that it is as if it never existed
            self.db.execute("UPDATE albums SET image=NULL WHERE id=?", (album_id,))
            # Always do this incase you remove the current tracks cover
            self.exaile.cover.set_image(NOCOVER_IMAGE)

    def set_parent(self, parent):
        """
            This is used to set the parent.
        """
        # Cause you cant replace self.parent
        self.mother = parent



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""
This file contains every miscellanious dialog and class that is not over
300 lines of code.  Once they read 300+ lines, they should be put into
their own file
"""

import urllib2, os, re, sys, threading, time, traceback, urllib, urlparse
from gettext import gettext as _
#import gtkmozembed

import pygtk, locale
pygtk.require('2.0')
import gtk, gobject, pango, gtk.glade
import cairo
from gtk.gdk import SCROLL_LEFT, SCROLL_RIGHT, SCROLL_UP, SCROLL_DOWN

try:
    from xl import mozembed
except ImportError:
    mozembed = None

try:
    import sexy
    SEXY_AVAIL = True
except ImportError:
    SEXY_AVAIL = False

import xl.path
import common, prefs

VALID_TAGS = (
    # Ogg Vorbis spec tags
    "title version album tracknumber artist genre performer copyright "
    "license organization description location contact isrc date "

    # Other tags we like
    "arranger author composer conductor lyricist discnumber labelid part "
    "website language encodedby bpm albumartist originaldate originalalbum "
    "originalartist recordingdate"
    ).split()

opener = urllib.FancyURLopener()
opener.addheaders.pop(0)
opener.addheader("User-Agent", "Mozilla")

def check_theme_file(file):
    """
        Checks a theme file to make sure it's compatible with the current
        version of Exaile.

        It will open the glade file, instantiate the "ThemeInfo" dialog, and
        check the "supported_versions" TextView for the current version.
        Supported versions are separated by a newline.  If the current version
        is not in the supported_versions, or the ThemeInfo dialog cannot be
        instantiated, this function will return False
    """
    
    current_version = sys.modules['__main__'].__version__
    try:
        xml = gtk.glade.XML(file, 'ThemeInfo', 'exaile')
        svtv = xml.get_widget('supported_versions')
        if not svtv: 
            log("supported_versions widget not found, not using theme")
            return False # the TextView could not be found

        supported_versions = svtv.get_buffer().get_text(
            svtv.get_buffer().get_start_iter(), 
            svtv.get_buffer().get_end_iter()).split('\n')

        if not current_version in supported_versions: 
            log("Current version: %s not found in %s, not using theme" %
                (current_version, supported_versions))
            return False

        log("%s theme is ok, continuing" % file)
        return True

    except:
        log_exception()
        return False


def glade_file(exaile):
    themefile = xl.path.get_config('exaile.extheme')
    if os.path.isfile(themefile):
        if not check_theme_file(themefile):
            return xl.path.get_data('exaile.glade')
        return themefile
    else:
        return xl.path.get_data('exaile.glade')

def get_default_encoding():
    """
        Returns the encoding to be used when dealing with file paths.  Do not
        use for other purposes.
    """
    #return 'utf-8'
    return sys.getfilesystemencoding() or sys.getdefaultencoding()

class ClearEntry(object):
    """
        A gtk.Entry with a clear icon
    """
    def __init__(self, change_func):
        """
            Initializes the entry
        """
        self.sexy = SEXY_AVAIL
        if self.sexy:
            self.entry = sexy.IconEntry()
            image = gtk.Image()
            image.set_from_stock('gtk-clear', gtk.ICON_SIZE_SMALL_TOOLBAR)
            self.entry.set_icon(sexy.ICON_ENTRY_SECONDARY, image)
            if change_func:
                self.entry.connect('icon-released', self.icon_released)
        else:
            self.entry = gtk.Entry()
        self.entry.connect('changed', change_func)

    def set_clear_callback(self, cb):
        """
            Sets the callback to be called after the clear button is pressed
        """
        self.clear_callback = cb

    def icon_released(self, *e):
        """
            Called when the user clicks the entry icon
        """
        self.entry.set_text('')

    def __getattr__(self, attr):
        """
            If this object doesn't have the attribute, check the gtk.Entry for
            it
        """
        if attr == 'entry': return self.entry
        return getattr(self.entry, attr)

def finish(repeat=True):
    """
        Waits for current pending gtk events to finish
    """
    while gtk.events_pending():
        gtk.main_iteration()
        if not repeat: break

class MmKeys:
    """
        Multimedia key grabber, using GNOME or the mmkeys module.

        Because of how the mmkeys module works, you must keep a reference to the
        instance of this class, specifically if the grabbing is done through
        mmkeys.
    """
    def __init__(self, application, callback):
        """
            Constructor.

            callback is a function that will be called with one of these as
            argument: 'Play', 'PlayPause', 'Pause', 'Stop', 'Previous', 'Next',
            or any future GNOME mmkey string.
        """
        self.application = application
        self.callback = callback

    def __del__(self):
        if self.use == 'gnome':
            try:
                self.__gnome.ReleaseMediaPlayerKeys(self.application)
            except:
                log_exception()

    def grab(self):
        """
            Try to grab multimedia keys.  Returns 'gnome', 'mmkeys', or None.
        """
        self.use = self.__use_gnome() or self.__use_mmkeys()
        return self.use

    def __use_gnome(self):
        def on_gnome_mmkey(app, key):
            if app == self.application:
                self.callback(key)

        try:
            import dbus
            bus = dbus.SessionBus()

            try:
                # new method (for gnome 2.22.x)
                obj = bus.get_object('org.gnome.SettingsDaemon',
                    '/org/gnome/SettingsDaemon/MediaKeys')
                self.__gnome = gnome = dbus.Interface(obj,
                    'org.gnome.SettingsDaemon.MediaKeys')
                gnome.GrabMediaPlayerKeys(self.application, 0)
                gnome.connect_to_signal('MediaPlayerKeyPressed', on_gnome_mmkey)
                log("Activated gnome mmkeys for gnome 2.22.x")
            except:
                # old method
                obj = bus.get_object('org.gnome.SettingsDaemon',
                    '/org/gnome/SettingsDaemon')
                self.__gnome = gnome = dbus.Interface(obj,
                    'org.gnome.SettingsDaemon')
                gnome.GrabMediaPlayerKeys(self.application, 0)
                gnome.connect_to_signal('MediaPlayerKeyPressed', on_gnome_mmkey)
                log("Activated gnome mmkeys for gnome 2.20.x")

        except:
            log_exception()
            return None

        return 'gnome'

    def __use_mmkeys(self):
        try:
            import mmkeys
        except ImportError:
            return None

        # Must keep a reference to the object.
        self.__mmkeys = keys = mmkeys.MmKeys()
        keys.connect('mm_playpause', lambda e, f: self.callback('PlayPause'))
        keys.connect('mm_stop', lambda e, f: self.callback('Stop'))
        keys.connect('mm_prev', lambda e, f: self.callback('Previous'))
        keys.connect('mm_next', lambda e, f: self.callback('Next'))

        return 'mmkeys'

class Menu(gtk.Menu):
    """
        A proxy for making it easier to add icons to menu items
    """
    def __init__(self):
        """
            Initializes the menu
        """
        gtk.Menu.__init__(self)

        self.show()

    def append_image(self, pixbuf, callback, data=None):
        """
            Appends a graphic as a menu item
        """
        item = gtk.MenuItem()
        image = gtk.Image()
        image.set_from_pixbuf(pixbuf)
        item.add(image)
        
        if callback: item.connect('activate', callback, data)
        gtk.Menu.append(self, item)
        item.show_all()
        return item

    def append(self, label, callback, stock_id=None, data=None):
        """
            Appends a menu item
        """
        if stock_id:
            item = gtk.MenuItem()
            hbox = gtk.HBox()
            hbox.set_spacing(5)
            item.add(hbox)
            image = gtk.image_new_from_stock(stock_id,
                gtk.ICON_SIZE_MENU)
            hbox.pack_start(image, False, True)
            label = gtk.Label(label)
            label.set_alignment(0, 0)
            hbox.pack_start(label, True, True)
        else:
            item = gtk.MenuItem(label)
            self.label = item.get_child()

        if callback: item.connect('activate', callback, data)
        gtk.Menu.append(self, item)
        item.show_all()
        return item

    def append_item(self, item):
        """
            Appends a menu item
        """
        gtk.Menu.append(self, item)
        item.show_all()

    def append_menu(self, label, menu, stock_id=None):
        """
            Appends a submenu
        """
        if stock_id:
            item = self.append(label, None, stock_id)
            item.set_submenu(menu)
            return item

        item = gtk.MenuItem(label)
        item.set_submenu(menu)
        item.show()
        gtk.Menu.append(self, item)

        return item

    def insert_menu(self, index, label, menu):
        """
            Inserts a menu at the specified index
        """
        item = gtk.MenuItem(label)
        item.set_submenu(menu)
        item.show()
        gtk.Menu.insert(self, item, index)

        return item

    def append_separator(self):
        """
            Adds a separator
        """
        item = gtk.SeparatorMenuItem()
        item.show()
        gtk.Menu.append(self, item)

class Adjustment(gtk.Adjustment):
    """
        Custom adjustment with convenience functions
    """
    def decrease(self, diff):
        self.props.value -= diff
    def increase(self, diff):
        self.props.value += diff
    def page_down(self):
        self.props.value -= self.props.page_increment
    def page_up(self):
        self.props.value += self.props.page_increment
    def step_down(self):
        self.props.value -= self.props.step_increment
    def step_up(self):
        self.props.value += self.props.step_increment

def get_icon(id, size=gtk.ICON_SIZE_BUTTON):
    """
        Returns a stock icon for the specified id and size
    """
    theme = gtk.icon_theme_get_default()
    try:
        icon = theme.load_icon(id, size, gtk.ICON_LOOKUP_NO_SVG)
        if icon: return icon
    except gobject.GError:
        pass
    
    return gtk.gdk.pixbuf_new_from_file(
        xl.path.get_data('images', 'default_theme', id + '.png'))


class ListBox(object):
    """
        Represents a list box
    """
    def __init__(self, widget, rows=None):
        """
            Initializes the widget
        """
        self.list = widget
        self.store = gtk.ListStore(gobject.TYPE_STRING)
        widget.set_headers_visible(False)
        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('', cell, text=0)
        self.list.append_column(col)
        self.rows = rows
        if not rows: self.rows = []
        
        if rows:
            for row in rows:
                self.store.append([row])

        self.list.set_model(self.store)

    def connect(self, signal, func, data=None):
        """
            Connects a signal to the underlying treeview
        """
        self.list.connect(signal, func, data)

    def append(self, row):
        """
            Appends a row to the list
        """
        self.rows.append(row)
        self.set_rows(self.rows)

    def remove(self, row):
        """
            Removes a row
        """
        try:
            index = self.rows.index(row)
        except ValueError:
            return
        path = (index,)
        iter = self.store.get_iter(path)
        self.store.remove(iter)
        del self.rows[index]

    def set_rows(self, rows):
        """
            Sets the rows
        """
        self.rows = rows
        self.store = gtk.ListStore(gobject.TYPE_STRING)
        for row in rows:
            self.store.append([row])

        self.list.set_model(self.store)

    def get_selection(self):
        """
            Returns the selection
        """
        selection = self.list.get_selection()
        (model, iter) = selection.get_selected()
        if not iter: return None
        return model.get_value(iter, 0)

# Reduce the notebook tabs' close button padding size.
gtk.rc_parse_string("""
    style "thinWidget" {
        xthickness = 0
        ythickness = 0
    }
    widget "*.tabCloseButton" style "thinWidget"
    """)

class NotebookTab(gtk.EventBox):
    """
        Shows a close image on a notebook tab
    """

    def __init__(self, exaile, title, page):
        gtk.EventBox.__init__(self)
        self.set_visible_window(False)
        self.connect('button_press_event', self.on_button_press)
        
        self.exaile = exaile
        self.title = title
        self.page = page
        self.nb = exaile.playlists_nb
        self.tips = gtk.Tooltips()
        
        self.hbox = hbox = gtk.HBox(False, 5)
        self.add(hbox)
        
        self.label = gtk.Label(title)
        hbox.pack_start(self.label, False, False)
        
        self.button = btn = gtk.Button()
        btn.set_name('tabCloseButton')
        btn.set_relief(gtk.RELIEF_NONE)
        btn.set_focus_on_click(False)
        btn.connect('clicked', self.do_close)
        btn.connect('button_press_event', self.on_button_press)
        self.tips.set_tip(btn, _("Close this tab"))
        image = gtk.Image()
        image.set_from_stock('gtk-close', gtk.ICON_SIZE_MENU)
        btn.add(image)
        hbox.pack_end(btn, False, False)
        
        self.show_all()

    def rename(self, widget, event):
        """
            Renames the tab
        """
        dialog = common.TextEntryDialog(self.exaile.window, 
            _("Enter the new name for this playlist"), _("Rename playlist"))
        if dialog.run() == gtk.RESPONSE_OK:
            name = dialog.get_value()
            self.title = name
            self.label.set_label(name)
        dialog.destroy()

    def save_playlist(self, widget, event):
        """
            Saves a playlist
        """
        songs = self.page.songs
        self.exaile.playlists_panel.on_add_playlist(widget, None, songs)

    def do_close(self, *args):
        """
            Closes the tab
        """
        self.exaile.close_page(self.page)

    def create_menu(self):
        """
            Creates the popup menu for this tab
        """
        menu = Menu()
        if self.page.type == 'track':
            menu.append(_("Rename"), self.rename)
        menu.append(_("Close"), self.do_close)

        if self.page.type == 'track':
            menu.append(_("Save Playlist"), self.save_playlist)
            
        self.menu = menu

    def on_button_press(self, widget, event):
        """
            Shows menu when user right-clicks on tab
        """
        if event.type == gtk.gdk._2BUTTON_PRESS:
            self.rename(widget, event)
            
        if event.button == 3:
            self.create_menu()
            self.menu.popup(None, None, None, event.button, event.time)
        return False

# Deprecated way for logging. Use the logger module instead.
from xl import logger
log = logger.log
log_multi = logger.log_multi
log_exception = logger.log_exception
log_stack = logger.log_stack

class URLFetcher(threading.Thread):
    """
        Fetches information from a URL
    """
    def __init__(self, server, path, done_func):
        """
            Initializes the thread
        """
        threading.Thread.__init__(self)
        self.server = server 
        self.path = path 

        logger.log(self.server + ' ' + self.path)
        self.setDaemon(True)
        self.done_func = done_func

    def run(self):
        """
            Called by thread.start()
        """
        try:
            response = urllib2.urlopen("http://" + self.server + self.path)
        except urllib2.URLError, e:
            self.notify("Could not load page. "
                "Error: %s" % e.reason)
            return

        except urllib2.HTTPError, e:
            self.notify("Could not load page. "
                "Error %d" % e.status)
            return

        page = response.read()
        self.notify(page)

    def notify(self, text):
        """
            Calls the done_func after the current gtk pending event
        """
        gobject.idle_add(self.done_func, self.server + self.path, text)

class BrowserWindow(gtk.VBox):
    """
        An HTML window for Wikipedia information
    """
    def __init__(self, exaile, url, nostyles=False):
        """
            Initializes the window
        """
        gtk.VBox.__init__(self)
        self.set_border_width(5)
        self.set_spacing(3)
        self.nostyles = nostyles
        self.action_count = 0

        self.exaile = exaile
        if not nostyles:
            top = gtk.HBox()
            top.set_spacing(3)

            self.back = gtk.Button()
            image = gtk.Image()
            image.set_from_stock('gtk-go-back', gtk.ICON_SIZE_SMALL_TOOLBAR)
            self.back.set_image(image)
            self.back.set_sensitive(False)
            self.back.connect('clicked', self.on_back)
            top.pack_start(self.back, False, False)

            self.next = gtk.Button()
            image = gtk.Image()
            image.set_from_stock('gtk-go-forward', gtk.ICON_SIZE_SMALL_TOOLBAR)
            self.next.set_image(image)
            self.next.connect('clicked', self.on_next)
            self.next.set_sensitive(False)
            top.pack_start(self.next, False, False)

            w = gtk.Button(_("Open Browser"))
            w.connect('clicked', self.on_open_browser)
            top.pack_start(w, False, False)

            self.entry = gtk.Entry()
            self.entry.connect('activate', self.entry_activate)
            top.pack_start(self.entry, True, True)
            self.pack_start(top, False, True)

        self.view = mozembed.MozClient()
        self.pack_start(self.view, True, True)
        if not nostyles:
            self.view.connect('location', self.on_location_change)

        self.show_all()
        finish()

        self.view.set_data('<html><body><b>' + _('Loading requested'
            ' information...') + '</b></body></html>', '')
        exaile.status.set_first(_('Loading page...'))
        self.view.connect('net-stop', self.on_net_stop)

        self.server = ''

        if url:
            self.load_url(url, self.action_count, False)

    def on_net_stop(self, *args):
        """
            Called when mozilla is done loading the page
        """
        self.exaile.status.set_first(None)

    def set_text(self, text):
        """
            Sets the text of the browser window

        """
        self.view.set_data(text, '')

    def entry_activate(self, *e):
        """
            Called when the user presses enter in the address bar
        """
        url = unicode(self.entry.get_text(), 'utf-8')
        self.load_url(url, self.action_count)

    def on_location_change(self, mozembed):
        # Only called when not self.nostyles
        self.entry.set_text(mozembed.get_location())
        self.back.set_sensitive(self.view.can_go_back())
        self.next.set_sensitive(self.view.can_go_forward())

    def on_next(self, widget):
        """
            Goes to the next entry in history
        """
        self.view.go_forward()
            
    def on_back(self, widget):
        """
            Goes to the previous entry in history
        """
        self.view.go_back()

    def on_open_browser(self, button):
        """
            Opens the current URL in a new browser window (if possible).
        """
        # This method is rarely used, so we only do the import when we need to.
        import webbrowser
        # "new=1" is to request new window.
        webbrowser.open(self.view.get_location(), new=1)

    def load_url(self, url, action_count, history=False):
        """
            Loads a URL, either from the cache, or from the website specified
        """
        self.view.load_url(url)

        if not self.nostyles:
            if self.view.can_go_back(): self.back.set_sensitive(True)
            if not self.view.can_go_forward(): self.next.set_sensitive(False)
            self.entry.set_sensitive(True)
            self.entry.set_text(url)

def show_about_dialog(parent, version):
    """
        Shows an about dialog
    """
    xml = gtk.glade.XML('exaile.glade', 'AboutDialog', 'exaile')
    dialog = xml.get_widget('AboutDialog')
    logo = gtk.gdk.pixbuf_new_from_file(xl.path.get_data('images',
        'exailelogo.png'))
    dialog.set_logo(logo)
    # HACK: GTK+ < 2.12 (2007-09-14) use set_name.
    try:
        dialog.set_program_name(_("Exaile"))
    except AttributeError:
        dialog.set_name(_("Exaile"))
    dialog.set_version("\n" + str(version))
    dialog.set_transient_for(parent)
    dialog.run()
    dialog.destroy()

class ImageWidget(gtk.Image):
    """
        Custom resizeable image widget
    """
    __gsignals__ = {
        'image-changed': (gobject.SIGNAL_RUN_LAST, 
            gobject.TYPE_NONE, (str,))
    }

    def __init__(self):
        """
            Initializes the image
        """
        gtk.Image.__init__(self)
        self.loc = ''

    def set_image_size(self, width, height):
        """
            Scales the size of the image
        """
        self.size = (width, height)

    def set_image(self, image, fill=False):
        """
            Sets the image
        """
        self.loc = image
        pixbuf = gtk.gdk.pixbuf_new_from_file(image)
        width, height = self.size
        if not fill:
            origw = float(pixbuf.get_width())
            origh = float(pixbuf.get_height())
            scale = min(width / origw, height / origh)
            width = int(origw * scale)
            height = int(origh * scale)
        self.width = width
        self.height = height
        scaled = pixbuf.scale_simple(width, height, gtk.gdk.INTERP_BILINEAR)
        self.set_from_pixbuf(scaled)

        scaled = pixbuf = None
        self.emit('image-changed', image)

class StatusBar(object):
    """
        A Custom status bar... allows for any component to be added to the
        status bar
    """
    def __init__(self, exaile):
        """
            Initializes the status bar
        """
        self.exaile = exaile
        self.xml = exaile.xml
        self.first_label = self.xml.get_widget('status_pos_1')
        self.second_label = self.xml.get_widget('status_pos_2')
        self.track_count_label = self.xml.get_widget('track_count_label')

    def set_track_count(self, count):
        """
            Sets the track count label
        """
        self.track_count_label.set_label(count)

    def clear(self):
        """
            Clears the text in the first section of the status bar
        """
        self.set_first(None)
        return False

    def set_first(self, text, time=0):
        """
            Sets the text in the first section of the status bar.  If time
            is specified, this text will be cleared after that amount of
            time (in milliseconds) has elapsed.
        """
        
        if not text:
            self.first_label.set_label('')

        else:
            self.first_label.set_label(text)

        if time:
            gobject.timeout_add(time, self.clear)

BITMAP_CACHE = dict()
def get_text_icon(widget, text, width, height, bgcolor='#456eac',   
    bordercolor=None):
    """
        Gets a bitmap icon with the specified text, width, and height
    """
    if BITMAP_CACHE.has_key("%s - %sx%s - %s" % (text, width, height, bgcolor)):
        return BITMAP_CACHE["%s - %sx%s - %s" % (text, width, height, bgcolor)]

    default_visual = gtk.gdk.visual_get_system()
    pixmap = gtk.gdk.Pixmap(None, width, height, default_visual.depth)
    colormap = gtk.gdk.colormap_get_system()
    white = colormap.alloc_color(65535, 65535, 65535)
    black = colormap.alloc_color(0, 0, 0)
    pixmap.set_colormap(colormap)
    gc = pixmap.new_gc(foreground=black, background=white)

    if not bordercolor: bordercolor = black
    else: 
        bordercolor = colormap.alloc_color(gtk.gdk.color_parse(bordercolor))
    gc.set_foreground(bordercolor)

    pixmap.draw_rectangle(gc, True, 0, 0, width, height)
    fg = colormap.alloc_color(gtk.gdk.color_parse(bgcolor))
    gc.set_foreground(fg)
    pixmap.draw_rectangle(gc, True, 1, 1, width - 2, height - 2)

    layout = widget.create_pango_layout(str(text))
    desc = pango.FontDescription("Sans 8")
    layout.set_font_description(desc)
    layout.set_alignment(pango.ALIGN_RIGHT)

    gc.set_foreground(white)
    inkRect, logicalRect = layout.get_pixel_extents()
    l, b, w, h = logicalRect
    x = ((width) / 2 - w / 2)
    y = ((height) / 2 - h / 2)
    pixmap.draw_layout(gc, x, y, layout)

    pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, False, 8, width, height)
    pixbuf = pixbuf.get_from_drawable(pixmap, colormap, 0, 0, 0,
        0, width, height)

    BITMAP_CACHE["%s - %sx%s - %s" % (text, width, height, bgcolor)] = pixbuf
    return pixbuf


class MiscTimer(object):
    """
        Creates a timer that executes a function
    """
    def __init__(self, func, time, runonce=False):
        """
            Expects the function to call.  If runonce is True, the
            timer only runs one time
        """
        self.func = func
        self.time = time
        self.runonce = runonce
        self.connect_id = None

    def start(self):
        """
            Starts the timer
        """
        self.connect_id = gobject.timeout_add(self.time, self.notify)

    def stop(self):
        """
            Stops the timer
        """
        if not self.connect_id: return
        gobject.source_remove(self.connect_id)
        self.connect_id = None

    def notify(self):
        """
            Called when the timer time has elapsed
        """
        self.func()
        if self.runonce: return False
        else: return True

class OSDWindow(object):
    """
        A popup window to show information on the current playing track
    """
    def __init__(self, exaile, settings, start_timer=True, draggable=False):
        """
            Initializes the popup
        """
        self.exaile = exaile
        self.draggable = draggable
        self.xml = gtk.glade.XML('exaile.glade', 'OSDWindow', 'exaile')
        self.window = self.xml.get_widget('OSDWindow')
        self.__timeout = None
        self.start_timer = start_timer

        self.color = gtk.gdk.color_parse(settings['osd/bgcolor'])
        self.settings = settings
        self.event = self.xml.get_widget('popup_event_box')
        self.box = self.xml.get_widget('image_box')

        self.progress = self.xml.get_widget('osd_progressbar')

        # Try to set the window opacity.  To do that we need the RGBA colormap,
        # which for some reason may not be available even if
        # Widget.is_composited is true.  In GTK+ >=2.12 all this can just be
        # modified to use Window.set_opacity.
        colormap = None
        # Widget.is_composited is only in GTK+ >=2.10.
        is_composited = getattr(self.window, 'is_composited', None)
        if is_composited and is_composited():
            screen = self.window.get_screen()
            colormap = screen.get_rgba_colormap()
        if colormap:
            self.window.set_colormap(colormap)
            self.window.set_app_paintable(True)
            self.window.connect("expose-event", self.expose_callback)

            self.progress.set_colormap(colormap)
            self.progress.set_app_paintable(True)
            self.progress.connect("expose-event", self.expose_callback)
        else:
            # Just set the background color in the old fashioned way
            self.window.modify_bg(gtk.STATE_NORMAL, self.color)
            self.progress.modify_bg(gtk.STATE_NORMAL, self.color)
            
        self.title = self.xml.get_widget('popup_title_label')

        self.window.set_size_request(settings['osd/w'], settings['osd/h'])
        self.cover = ImageWidget()
        self.box.pack_start(self.cover, False, False)
        self.window.move(settings['osd/x'], settings['osd/y'])
        self.cover.set_image_size(settings['osd/h'] - 8, settings['osd/h'] - 8)
        self.event.connect('button_press_event', self.start_dragging)
        self.event.connect('button_release_event', self.stop_dragging)
        self.exaile.connect('timer_update', self.timer_update)
        self.__handler = None

    def expose_callback(self, widget, event):
        cr = widget.window.cairo_create()
        cr.set_operator(cairo.OPERATOR_SOURCE)

        cr.rectangle(event.area.x, event.area.y,
            event.area.width, event.area.height)
        cr.clip()

        opacity = self.settings['osd/opacity']
        
        cr.set_source_rgba(self.color.red/65535.0, self.color.green/65535.0, 
                self.color.blue/65535.0, opacity/100.0)
        cr.paint()
        return False

    def timer_update(self, *e):
        if self.window.get_property('visible'):
            self.progress.set_fraction(self.exaile.new_progressbar.get_fraction())

    def expose_callback(self, widget, event):
        cr = widget.window.cairo_create()
        cr.set_operator(cairo.OPERATOR_SOURCE)

        cr.rectangle(event.area.x, event.area.y,
            event.area.width, event.area.height)
        cr.clip()

        opacity = self.settings['osd/opacity']
        
        cr.set_source_rgba(self.color.red/65535.0, self.color.green/65535.0, 
                self.color.blue/65535.0, opacity/100.0)
        cr.paint()
        return False


    def start_dragging(self, widget, event):
        """
            Called when the user starts dragging the window
        """
        if not self.draggable:
            self.window.hide()
            return
        self.__start = event.x, event.y
        self.__handler = self.window.connect('motion_notify_event',
            self.dragging)
        if self.__timeout: gobject.source_remove(self.__timeout)
        self.__timeout = None

    def stop_dragging(self, widget, event):
        """
            Called when the user stops dragging the mouse
        """
        global POPUP
        if self.__handler: self.window.disconnect(self.__handler)
        self.__handler = None
        if self.start_timer:
            self.__timeout = gobject.timeout_add(4000, self.window.hide)
        settings = self.exaile.settings
        (w, h) = self.window.get_size()
        (x, y) = self.window.get_position()

        settings['osd/x'] = x
        settings['osd/y'] = y
        settings['osd/h'] = h
        settings['osd/w'] = w
    
        POPUP = OSDWindow(self.exaile, get_osd_settings(settings))

    def dragging(self, widget, event):
        """
            Called when the user drags the window
        """
        self.window.move(int(event.x_root - self.__start[0]),
            int(event.y_root - self.__start[1]))

    def show_track_osd(self, track, text, cover):
        """
            Shows a popup specific to a track
        """
        self.progress.show()
        text = text.replace("&", "&amp;") # we don't allow entity refs here

        for item in ('title', 'artist', 'album', 'length', 'track', 'bitrate',
            'genre', 'year', 'rating'):
            try:
                value = getattr(track, item)
            except AttributeError:
                continue
            if not isinstance(value, basestring):
                value = unicode(value)
            text = text.replace("{%s}" % item, common.escape_xml(value))

        text = text.replace("{volume}", "%d%%" %
            self.exaile.get_volume_percent())
        text = text.replace("\\{", "{")
        text = text.replace("\\}", "}")
        self.show_osd(text, cover, True)

    def show_osd(self, title, cover, show_progress=False):
        """
            Displays the popup for 4 seconds
        """
        self.progress.hide()
        if self.__timeout:
            gobject.source_remove(self.__timeout)
            self.window.hide()
        
        text = "<span font_desc='%s' foreground='%s'>%s</span>" % \
            (self.settings['osd/text_font'], self.settings['osd/text_color'],
            title)
        self.title.set_markup(text)

        if not cover:
            cover = xl.path.get_data('images', 'nocover.png')

        self.cover.set_image(cover)
        self.window.show_all()
        if not show_progress:
            self.progress.hide()

        if self.start_timer: 
            self.__timeout = gobject.timeout_add(4000, self.window.hide)

POPUP = None
def get_osd(exaile, settings):
    """
        Gets a popup instance
    """
    global POPUP
    
    if not POPUP:
        POPUP = OSDWindow(exaile, settings)

    return POPUP

def get_osd_settings(settings):
    info = dict()
    info['osd/bgcolor'] = settings.get_str("osd/bgcolor", "#567ea2")
    info['osd/opacity'] = settings.get_int("osd/opacity", 80)
    info['osd/w'] = settings.get_int("osd/w", 400)
    info['osd/h'] = settings.get_int("osd/h", 95)
    info['osd/y'] = settings.get_int("osd/y", 0)
    info['osd/x'] = settings.get_int("osd/x", 0)
    info['osd/display_text'] = settings.get_str('osd/display_text', 
        prefs.TEXT_VIEW_DEFAULT)
    info['osd/text_font'] = settings.get_str('osd/text_font',
        'Sans 10')
    info['osd/text_color'] = settings.get_str('osd/text_color',
        '#ffffff')

    return info

class DragTreeView(gtk.TreeView):
    """
        A TextView that does easy dragging/selecting/popup menu
    """
    def __init__(self, cont, receive=True, source=True):
        """
            Initializes the tree and sets up the various callbacks
        """
        gtk.TreeView.__init__(self)
        self.cont = cont

        self.targets = [("text/uri-list", 0, 0)]

        if source:
            self.drag_source_set(
                gtk.gdk.BUTTON1_MASK, self.targets,
                gtk.gdk.ACTION_COPY|gtk.gdk.ACTION_MOVE)

        if receive:
            self.drag_dest_set(gtk.DEST_DEFAULT_ALL, self.targets, 
                gtk.gdk.ACTION_COPY|gtk.gdk.ACTION_DEFAULT)
            self.connect('drag_data_received', 
                self.cont.drag_data_received)
        self.receive = receive
        self.dragging = False
        self.connect('drag_begin', self.drag_begin)
        self.connect('drag_end', self.drag_end)
        self.connect('drag_motion', self.drag_motion)
        self.connect('button_release_event', self.button_release)
        self.connect('button_press_event', self.button_press)

        if source:
            self.connect('drag_data_get', self.cont.drag_get_data)
            self.drag_source_set_icon_stock('gtk-dnd')

    def button_release(self, button, event):
        """
            Called when a button is released
        """
        if event.button != 1 or self.dragging: 
            self.dragging = False
            return True
        if event.state & (gtk.gdk.SHIFT_MASK|gtk.gdk.CONTROL_MASK):
            return True
        selection = self.get_selection()
        x, y = event.get_coords()
        x = int(x); y = int(y)

        path = self.get_path_at_pos(x, y)
        if not path: return False
        selection.unselect_all()
        selection.select_path(path[0])

    def drag_end(self, list, context):
        """
            Called when the dnd is ended
        """
        self.dragging = False
        self.unset_rows_drag_dest()
        self.drag_dest_set(gtk.DEST_DEFAULT_ALL, self.targets, 
            gtk.gdk.ACTION_COPY|gtk.gdk.ACTION_MOVE)

    def drag_begin(self, list, context):
        """
            Called when dnd is started
        """
        self.dragging = True

        context.drag_abort(gtk.get_current_event_time())
        selection = self.get_selection()
        if selection.count_selected_rows() > 1:
            self.drag_source_set_icon_stock('gtk-dnd-multiple')
        else: self.drag_source_set_icon_stock('gtk-dnd')
        return False

    def drag_motion(self, treeview, context, x, y, timestamp):
        """
            Called when a row is dragged over this treeview
        """
        if not self.receive:
            return
        self.enable_model_drag_dest(self.targets,
            gtk.gdk.ACTION_DEFAULT)
        info = treeview.get_dest_row_at_pos(x, y)
        if not info: return
        treeview.set_drag_dest_row(info[0], info[1])

    def button_press(self, button, event):
        """
            The popup menu that is displayed when you right click in the
            playlist
        """
        selection = self.get_selection()
        (x, y) = event.get_coords()
        x = int(x)
        y = int(y)
        path = self.get_path_at_pos(x, y)
        if not path: return True
            
        if event.button != 3: 
            if event.type == gtk.gdk._2BUTTON_PRESS:
                self.cont.button_press(button, event)

            if selection.count_selected_rows() <= 1: 
                return False
            else: 
                if selection.path_is_selected(path[0]): 
                    if event.state & (gtk.gdk.SHIFT_MASK|gtk.gdk.CONTROL_MASK):
                        selection.unselect_path(path[0])
                    return True
                elif not event.state & (gtk.gdk.SHIFT_MASK|gtk.gdk.CONTROL_MASK):
                    return True
                return False

        if not selection.count_selected_rows():
            selection.select_path(path[0])
        return self.cont.button_press(button, event)

PLAYLIST_EXTS = []

class PlaylistParser(object):
    def __init__(self, name, url=None):
        self.name = name
        self.url = []

        self.parse_file(url)

    def add_url(self, url, title=None, album=''):
        if not title: title=url
        try: 
            url = list(urlparse.urlsplit(url))
        except:
            pass

        if not url[0]:
            url[0] = 'file'

        item = {
            'url': url,
            'title': title,
            'album': album
        }
        self.url.append(item)

    def get_urls(self):
        return self.url

    def get_full(self):
        return [urlparse.urlunsplit(u) for u in self.url]

    def get_name(self):
        return self.name

    def parse_file(self, url):
        f = urllib.urlopen(url)
        try:
            return self._do_parse_file(f)
        finally:
            try:
                f.close()
            except:
                pass

    def _do_parse_file(self, file):
        raise NotImplementedError

    def _get_url_from_path(self, basedir, s):
        if not urlparse.urlsplit(s)[0]: # no scheme --> local path
            if not os.path.isabs(s): # relative path
                s = os.path.join(basedir, s)
            s = 'file://' + urllib.quote(s)
        return s

class M3UParser(PlaylistParser):
    PLAYLIST_EXTS.append('.m3u')
    REGEX = re.compile(r'#EXTINF:-?\d+,(.*?)[\r\n]+([^\r\n]+)', re.DOTALL) 

    def _do_parse_file(self, file):
        # Read first line to see if this is extended M3U.
        firstline = file.readline()
        if firstline.strip() == "#EXTM3U":
            return self._do_parse_extended(file)

        basedir = os.path.dirname(file.url)
        for line in [firstline] + file.readlines():
            line = line.strip()
            if line and not line.startswith("#"):
                url = self._get_url_from_path(basedir, line)
                self.add_url(url)

        return True

    def _do_parse_extended(self, file):
        data = file.read()
        items = self.REGEX.findall(data)
        if items:
            basedir = os.path.dirname(file.url)
            for item in items:
                url = self._get_url_from_path(basedir, item[1])
                self.add_url(url, title=item[0], album=url)

        return True

class PlsParser(PlaylistParser):
    PLAYLIST_EXTS.append('.pls')
    REGEX = re.compile(r'[fF]ile(\d+)=(.*?)\n[tT]itle(\1)=(.*?)\n', re.DOTALL)

    def _do_parse_file(self, file):
        data = file.read()
        items = self.REGEX.findall(data)

        if items:
            basedir = os.path.dirname(file.url)
            for item in items:
                url = self._get_url_from_path(basedir, item[1])
                self.add_url(url, title=url, album=item[3])

        return True
   
class ASXParser(PlaylistParser):
    PLAYLIST_EXTS.append('.asx')
    REGEX = re.compile(r'''<ref\s+href\s*=\s*(['"])(.*?)\1''',
        re.DOTALL | re.IGNORECASE | re.MULTILINE)

    def _do_parse_file(self, file):
        data = file.read()
        items = self.REGEX.findall(data)

        if items:
            basedir = os.path.dirname(file.url)
            for item in items:
                url = self._get_url_from_path(basedir, item[1])
                self.add_url(url)

        return True


class XSPFParser(PlaylistParser):
    PLAYLIST_EXTS.append('.xspf')

    def _do_parse_file(self, file):
        try:
            import xml.etree.cElementTree as cETree
        except:
            return True

        tree = cETree.ElementTree(file=file)
        ns = "{http://xspf.org/ns/0/}"
        tracks = tree.find("%strackList"%ns).findall("%strack"%ns)
        for t in tracks:
            url = t.find("%slocation"%ns).text.strip()
            self.add_url(url)

        return True


# EXCEPTIONS

class IsFileException(Exception):
    """
        Thrown when a TrackGroup contains one or more tracks that aren't files
    """
    pass

class TagUnsupportedException(Exception):
    """
        Thrown when a TrackGroup has a file that doesn't 
        support a particular tag
    """
    def __init__(self, tag):
        self.tag = tag





# Description:
# wmainfo-py gives you access to low level information on wma files.
# * It identifies all "ASF_..." objects and shows each objects size
# * It returns info such as bitrate, size, length, creation date etc
# * It returns meta-tags from ASF_Content_Description_Object
#
# Note:
# I wrestled with the ASF spec (150 page .doc format!) with no joy for
# a while, then found Dan Sully's Audio-WMA Perl module:
# (http://cpants.perl.org/dist/Audio-WMA :: http://www.slimdevices.com/)
# This entire library is essentially a translation of parts of WMA.pm
# to Python. All credit for the hard work is owed to Dan...
#
# License:: Artistic/Perl
# Author:: Darren Kirby (mailto:bulliver@badcomputer.org)
# Website:: http://badcomputer.org/unix/code/wmainfo//

from os import stat
from struct import unpack
import time, re

class WmaInfoError:
    def __init__(self, error):
        self.error = error
    def __str__(self):
        return self.error

class WmaInfo:
    def __init__(self, file, debug=None):
        # 'public' attributes
        self.drm = None         # boolean
        self.tags = {}          # hash
        self.headerObject = {}  # hash of arrays
        self.info = {}          # hash

        self.file = file
        self.debug = debug
        self.__parseWmaHeader()

    def printobjects(self):
        '''
         ASF_Header_Object prints: "name: GUID size num_objects"
         All other objects print: "name: GUID size offset"
        '''
        for key,val in self.headerObject.iteritems():
            print "%s: %s %i %i" % (key, val[0], val[1], val[2])

    def hasdrm(self):
        '''
        Returns true if the file has DRM, ie: if a
        *Content_Encryption_Object is present
        '''
        if self.drm:
            return True
        else:
            return False

    def printtags(self):
        '''
        This is all the 'id3' like info gathered from:
        ASF_Content_Description_Object and
        ASF_Extended_Content_Description_Object
        '''
        for k,v in self.tags.iteritems(): 
            print "%s:%s%s" % (k, " " * (13 - len(k)), v)

    def hastag(self, tag):
        '''returns true if tags[tag] has a value'''
        if self.tags.has_key(tag) and self.tags[tag] != "":
            return True
        else:
            return False

    def printinfo(self):
        '''
        This is all the 'non-id3' like info gathered from
        ASF_File_Properties_Object and
        ASF_Extended_Content_Description_Object
        '''
        for k,v in self.info.iteritems():
            print "%s:%s%s" % (k, " " * (20 - len(k)), v)

    def hasinfo(self, field):
        '''Returns true if info[field] has a value'''
        if self.info.has_key(field) and self.tags[tag] != "":
            return True
        else:
            return False

    def parsestream(self):
        '''
        I don't think most people will want/need this info
        so it is not parsed automatically
        '''
        self.stream = {}
        try:
            offset = int(self.headerObject['ASF_Stream_Properties_Object'][2])
            self.__parseASFStreamPropertiesObject(offset)
        except:
            raise WmaInfoError("Cannot grok ASF_Stream_Properties_Object")

    def __parseWmaHeader(self):
        self.size = int(stat(self.file)[6])
        self.fh = open(self.file, "rb")
        self.offset = 0
        self.fileOffset = 30
        self.guidMapping = self.__knownGUIDs()
        self.reverseGuidMapping = {}
        for k,v in self.guidMapping.iteritems():
            self.reverseGuidMapping[v] = k

        try:
            objectId      = self.__byteStringToGUID(self.fh.read(16))
            objectSize    = unpack("<2L", self.fh.read(8))[0]
            headerObjects = unpack("<1L", self.fh.read(4))[0]
            reserved1     = unpack("b", self.fh.read(1))[0]
            reserved2     = unpack("b", self.fh.read(1))[0]
            objectIdName  = self.reverseGuidMapping[objectId]
        except:
            raise WmaInfoError(self.file + " doesn't appear to have a valid ASF header")

        if objectSize > self.size:
            raise WmaInfoError("Header size reported larger than file size")

        self.headerObject[objectIdName] = [objectId,  objectSize, headerObjects, reserved1, reserved2]

        if self.debug:
            print "objectId:      %s" % objectId
            print "objectIdName:  %s" % self.reverseGuidMapping[objectId]
            print "objectSize:    %s" % objectSize
            print "headerObjects: %s" % headerObjects
            print "reserved1:     %s" % reserved1
            print "reserved2:     %s" % reserved2

        self.headerData = self.fh.read(objectSize - 30)
        self.fh.close # Done with the file before we get going

        for n in range(headerObjects):
            nextObject     = self.__readAndIncrementOffset(16)
            nextObjectText = self.__byteStringToGUID(nextObject)
            nextObjectSize = self.__parse64BitString(self.__readAndIncrementOffset(8))
            nextObjectName = self.reverseGuidMapping[nextObjectText]

            self.headerObject[nextObjectName] = [nextObjectText, nextObjectSize, self.fileOffset]
            self.fileOffset += nextObjectSize

            if self.debug:
                print "nextObjectGUID: %s" % nextObjectText
                print "nextObjectName: %s" % nextObjectName
                print "nextObjectSize: %s" % nextObjectSize

            # start looking at object contents
            if (nextObjectName == 'ASF_File_Properties_Object'):
                self.__parseASFFilePropertiesObject()
                continue
            elif (nextObjectName == 'ASF_Content_Description_Object'):
                self.__parseASFContentDescriptionObject()
                continue
            elif (nextObjectName == 'ASF_Extended_Content_Description_Object'):
                self.__parseASFExtendedContentDescriptionObject()
                continue
            elif (nextObjectName == 'ASF_Content_Encryption_Object') or (nextObjectName == 'ASF_Extended_Content_Encryption_Object'):
                self.drm = 1

            # set our next object size
            self.offset += nextObjectSize - 24

    def __parseASFFilePropertiesObject(self):
        fileid                          = self.__readAndIncrementOffset(16)
        self.info['fileid_guid']        = self.__byteStringToGUID(fileid)
        self.info['filesize']           = int(self.__parse64BitString(self.__readAndIncrementOffset(8)))
        self.info['creation_date']      = unpack("<Q", self.__readAndIncrementOffset(8))[0]
        self.info['creation_date_unix'] = self.__fileTimeToUnixTime(self.info['creation_date'])
        self.info['creation_string']    = time.strftime("%c", time.gmtime(self.info['creation_date_unix']))
        self.info['data_packets']       = unpack("<4H", self.__readAndIncrementOffset(8))[0]
        self.info['play_duration']      = self.__parse64BitString(self.__readAndIncrementOffset(8))
        self.info['send_duration']      = self.__parse64BitString(self.__readAndIncrementOffset(8))
        self.info['preroll']            = unpack("<4H", self.__readAndIncrementOffset(8))[0]
        self.info['playtime_seconds']   = int(self.info['play_duration'] / 10000000 - self.info['preroll'] / 1000)
        flags_raw                       = unpack("<L", self.__readAndIncrementOffset(4))[0]
        if (flags_raw & 0x0001) == False:
            self.info['broadcast'] = 0
        else:
            self.info['broadcast'] = 1
        if (flags_raw & 0x0002) == False:
            self.info['seekable'] = 0
        else:
            self.info['seekable'] = 1
        self.info['min_packet_size']    = unpack("<2H", self.__readAndIncrementOffset(4))[0]
        self.info['max_packet_size']    = unpack("<2H", self.__readAndIncrementOffset(4))[0]
        self.info['max_bitrate']        = unpack("<2H", self.__readAndIncrementOffset(4))[0]
        self.info['bitrate']            = self.info['max_bitrate'] / 1000

        if self.debug:
            for key,val in self.info.iteritems():
                print "%s: %s" % (key, val)

    def __parseASFContentDescriptionObject(self):
        lengths = {}
        keys = ["Title", "Author", "Copyright", "Description", "Rating"]
        for key in keys:  # read the lengths of each key
            lengths[key] = unpack("H", self.__readAndIncrementOffset(2))[0]
        for key in keys:  # now pull the data based on length
            self.tags[key] = self.__decodeBinaryString(self.__readAndIncrementOffset(lengths[key]))

    def __parseASFExtendedContentDescriptionObject(self):
        ext_info = {}
        ext_info['content_count'] = unpack("H", self.__readAndIncrementOffset(2))[0]
        for n in range(ext_info['content_count']):
            ext = {}
            ext['base_offset']  = self.offset + 30
            ext['name_length']  = unpack("H", self.__readAndIncrementOffset(2))[0]
            ext['name']         = self.__decodeBinaryString(self.__readAndIncrementOffset(ext['name_length']))
            ext['value_type']   = unpack("H", self.__readAndIncrementOffset(2))[0]
            ext['value_length'] = unpack("H", self.__readAndIncrementOffset(2))[0]

            value = self.__readAndIncrementOffset(ext['value_length'])
            if ext['value_type'] <= 1:
                ext['value'] = self.__decodeBinaryString(value)
            elif ext['value_type'] == 4:
                ext['value'] = self.__parse64BitString(value)
            else:
                valTypeTemplates = ["", "", "<2H", "<2H", "", "H"]
                ext['value'] = unpack(valTypeTemplates[ext['value_type']], value)[0]

            if self.debug:
                print "base_offset:  %i" % ext['base_offset']
                print "name length:  %i" % ext['name_length']
                print "name:         %s" % ext['name']
                print "value type:   %i" % ext['value_type']
                print "value length: %i" % ext['value_length']
                print "value:        %s" % ext['value']

            ext_info[ext['name']] = ext['value']

        # Sort and dispatch info
        strMatch = re.compile("(TrackNumber|AlbumTitle|AlbumArtist|Genre|Year|Composer|Mood|Lyrics|BeatsPerMinute)")
        for k,v in ext_info.iteritems():
            if strMatch.search(k) != None:
                self.tags[k.replace("WM/", "")] = v # dump "WM/"
            else:
                self.info[k.replace("WM/", "")] = v

    def __parseASFStreamPropertiesObject(self, offset):
        self.offset = offset - 6 # gained an extra 6 bytes somewhere?!

        streamType                        = self.__readAndIncrementOffset(16)
        self.stream['stream_type_guid']   = self.__byteStringToGUID(streamType)
        self.stream['stream_type_name']   = self.reverseGuidMapping[self.stream['stream_type_guid']]
        errorType                         = self.__readAndIncrementOffset(16)
        self.stream['error_correct_guid'] = self.__byteStringToGUID(errorType)
        self.stream['error_correct_name'] = self.reverseGuidMapping[self.stream['error_correct_guid']]

        self.stream['time_offset']        = unpack('4H', self.__readAndIncrementOffset(8))[0]
        self.stream['type_data_length']   = unpack('2H', self.__readAndIncrementOffset(4))[0]
        self.stream['error_data_length']  = unpack('2H', self.__readAndIncrementOffset(4))[0]
        flags_raw                         = unpack('H', self.__readAndIncrementOffset(2))[0]
        self.stream['stream_number']      = flags_raw & 0x007F
        self.stream['encrypted']          = flags_raw & 0x8000

        #  reserved - set to zero
        self.__readAndIncrementOffset(4)

        self.stream['type_specific_data'] = self.__readAndIncrementOffset(self.stream['type_data_length'])
        self.stream['error_correct_data'] = self.__readAndIncrementOffset(self.stream['error_data_length'])

        if (self.stream['stream_type_name'] == 'ASF_Audio_Media'):
            self.__parseASFAudioMediaObject()

    def __parseASFAudioMediaObject(self):
        data = self.stream['type_specific_data'][0:16]
        self.stream['audio_channels']        = unpack('H', data[2:4])[0]
        self.stream['audio_sample_rate']     = unpack('2H', data[4:8])[0]
        self.stream['audio_bitrate']         = unpack('2H', data[8:12])[0] * 8
        self.stream['audio_bits_per_sample'] = unpack('H', data[14:16])[0]

    def __decodeBinaryString(self, data):
        textString = data.decode('utf-16le','ignore')[0:-1]
        return textString

    def __readAndIncrementOffset(self, size):
        value = self.headerData[self.offset:(self.offset + size)]
        self.offset += size
        return value

    def __byteStringToGUID(self, byteString):
        byteString = unpack("16B" , byteString)
        guidString  = "%02X" % byteString[3]
        guidString += "%02X" % byteString[2]
        guidString += "%02X" % byteString[1]
        guidString += "%02X" % byteString[0]
        guidString += '-'
        guidString += "%02X" % byteString[5]
        guidString += "%02X" % byteString[4]
        guidString += '-'
        guidString += "%02X" % byteString[7]
        guidString += "%02X" % byteString[6]
        guidString += '-'
        guidString += "%02X" % byteString[8]
        guidString += "%02X" % byteString[9]
        guidString += '-'
        guidString += "%02X" % byteString[10]
        guidString += "%02X" % byteString[11]
        guidString += "%02X" % byteString[12]
        guidString += "%02X" % byteString[13]
        guidString += "%02X" % byteString[14]
        guidString += "%02X" % byteString[15]
        return guidString

    def __parse64BitString(self, data):
        d = unpack('<2L', data)
        return d[1] * 2 ** 32 + d[0]

    def __fileTimeToUnixTime(self, time):
        return int((time - 116444736000000000) / 10000000)

    def __knownGUIDs(self):
        guidMapping = {
            'ASF_Extended_Stream_Properties_Object'   : '14E6A5CB-C672-4332-8399-A96952065B5A',
            'ASF_Padding_Object'                      : '1806D474-CADF-4509-A4BA-9AABCB96AAE8',
            'ASF_Payload_Ext_Syst_Pixel_Aspect_Ratio' : '1B1EE554-F9EA-4BC8-821A-376B74E4C4B8',
            'ASF_Script_Command_Object'               : '1EFB1A30-0B62-11D0-A39B-00A0C90348F6',
            'ASF_No_Error_Correction'                 : '20FB5700-5B55-11CF-A8FD-00805F5C442B',
            'ASF_Content_Branding_Object'             : '2211B3FA-BD23-11D2-B4B7-00A0C955FC6E',
            'ASF_Content_Encryption_Object'           : '2211B3FB-BD23-11D2-B4B7-00A0C955FC6E',
            'ASF_Digital_Signature_Object'            : '2211B3FC-BD23-11D2-B4B7-00A0C955FC6E',
            'ASF_Extended_Content_Encryption_Object'  : '298AE614-2622-4C17-B935-DAE07EE9289C',
            'ASF_Simple_Index_Object'                 : '33000890-E5B1-11CF-89F4-00A0C90349CB',
            'ASF_Degradable_JPEG_Media'               : '35907DE0-E415-11CF-A917-00805F5C442B',
            'ASF_Payload_Extension_System_Timecode'   : '399595EC-8667-4E2D-8FDB-98814CE76C1E',
            'ASF_Binary_Media'                        : '3AFB65E2-47EF-40F2-AC2C-70A90D71D343',
            'ASF_Timecode_Index_Object'               : '3CB73FD0-0C4A-4803-953D-EDF7B6228F0C',
            'ASF_Metadata_Library_Object'             : '44231C94-9498-49D1-A141-1D134E457054',
            'ASF_Reserved_3'                          : '4B1ACBE3-100B-11D0-A39B-00A0C90348F6',
            'ASF_Reserved_4'                          : '4CFEDB20-75F6-11CF-9C0F-00A0C90349CB',
            'ASF_Command_Media'                       : '59DACFC0-59E6-11D0-A3AC-00A0C90348F6',
            'ASF_Header_Extension_Object'             : '5FBF03B5-A92E-11CF-8EE3-00C00C205365',
            'ASF_Media_Object_Index_Parameters_Obj'   : '6B203BAD-3F11-4E84-ACA8-D7613DE2CFA7',
            'ASF_Header_Object'                       : '75B22630-668E-11CF-A6D9-00AA0062CE6C',
            'ASF_Content_Description_Object'          : '75B22633-668E-11CF-A6D9-00AA0062CE6C',
            'ASF_Error_Correction_Object'             : '75B22635-668E-11CF-A6D9-00AA0062CE6C',
            'ASF_Data_Object'                         : '75B22636-668E-11CF-A6D9-00AA0062CE6C',
            'ASF_Web_Stream_Media_Subtype'            : '776257D4-C627-41CB-8F81-7AC7FF1C40CC',
            'ASF_Stream_Bitrate_Properties_Object'    : '7BF875CE-468D-11D1-8D82-006097C9A2B2',
            'ASF_Language_List_Object'                : '7C4346A9-EFE0-4BFC-B229-393EDE415C85',
            'ASF_Codec_List_Object'                   : '86D15240-311D-11D0-A3A4-00A0C90348F6',
            'ASF_Reserved_2'                          : '86D15241-311D-11D0-A3A4-00A0C90348F6',
            'ASF_File_Properties_Object'              : '8CABDCA1-A947-11CF-8EE4-00C00C205365',
            'ASF_File_Transfer_Media'                 : '91BD222C-F21C-497A-8B6D-5AA86BFC0185',
            'ASF_Advanced_Mutual_Exclusion_Object'    : 'A08649CF-4775-4670-8A16-6E35357566CD',
            'ASF_Bandwidth_Sharing_Object'            : 'A69609E6-517B-11D2-B6AF-00C04FD908E9',
            'ASF_Reserved_1'                          : 'ABD3D211-A9BA-11CF-8EE6-00C00C205365',
            'ASF_Bandwidth_Sharing_Exclusive'         : 'AF6060AA-5197-11D2-B6AF-00C04FD908E9',
            'ASF_Bandwidth_Sharing_Partial'           : 'AF6060AB-5197-11D2-B6AF-00C04FD908E9',
            'ASF_JFIF_Media'                          : 'B61BE100-5B4E-11CF-A8FD-00805F5C442B',
            'ASF_Stream_Properties_Object'            : 'B7DC0791-A9B7-11CF-8EE6-00C00C205365',
            'ASF_Video_Media'                         : 'BC19EFC0-5B4D-11CF-A8FD-00805F5C442B',
            'ASF_Audio_Spread'                        : 'BFC3CD50-618F-11CF-8BB2-00AA00B4E220',
            'ASF_Metadata_Object'                     : 'C5F8CBEA-5BAF-4877-8467-AA8C44FA4CCA',
            'ASF_Payload_Ext_Syst_Sample_Duration'    : 'C6BD9450-867F-4907-83A3-C77921B733AD',
            'ASF_Group_Mutual_Exclusion_Object'       : 'D1465A40-5A79-4338-B71B-E36B8FD6C249',
            'ASF_Extended_Content_Description_Object' : 'D2D0A440-E307-11D2-97F0-00A0C95EA850',
            'ASF_Stream_Prioritization_Object'        : 'D4FED15B-88D3-454F-81F0-ED5C45999E24',
            'ASF_Payload_Ext_System_Content_Type'     : 'D590DC20-07BC-436C-9CF7-F3BBFBF1A4DC',
            'ASF_Index_Object'                        : 'D6E229D3-35DA-11D1-9034-00A0C90349BE',
            'ASF_Bitrate_Mutual_Exclusion_Object'     : 'D6E229DC-35DA-11D1-9034-00A0C90349BE',
            'ASF_Index_Parameters_Object'             : 'D6E229DF-35DA-11D1-9034-00A0C90349BE',
            'ASF_Mutex_Language'                      : 'D6E22A00-35DA-11D1-9034-00A0C90349BE',
            'ASF_Mutex_Bitrate'                       : 'D6E22A01-35DA-11D1-9034-00A0C90349BE',
            'ASF_Mutex_Unknown'                       : 'D6E22A02-35DA-11D1-9034-00A0C90349BE',
            'ASF_Web_Stream_Format'                   : 'DA1E6B13-8359-4050-B398-388E965BF00C',
            'ASF_Payload_Ext_System_File_Name'        : 'E165EC0E-19ED-45D7-B4A7-25CBD1E28E9B',
            'ASF_Marker_Object'                       : 'F487CD01-A951-11CF-8EE6-00C00C205365',
            'ASF_Timecode_Index_Parameters_Object'    : 'F55E496D-9797-4B5D-8C8B-604DFE9BFB24',
            'ASF_Audio_Media'                         : 'F8699E40-5B4D-11CF-A8FD-00805F5C442B',
            'ASF_Media_Object_Index_Object'           : 'FEB103F8-12AD-4C64-840F-2A1D2F7AD48C',
            'ASF_Alt_Extended_Content_Encryption_Obj' : 'FF889EF1-ADEE-40DA-9E71-98704BB928CE',
        }
        return guidMapping

if __name__ == '__main__':
    import sys
    foo = WmaInfo(sys.argv[1])
    print "### Info: ###\n"
    foo.printinfo()
    print
    print "### Tags: ###\n"
    foo.printtags()
    print
    print "### Objects ###\n"
    foo.printobjects()
    sys.exit(0)




from xl import xlmisc

TYPE = 'wav'

def fill_tag_from_path(tr):
    tr.tags['title'] = tr.loc.encode(xlmisc.get_default_encoding())

def can_change(tag):
    return False

def is_multi():
    return False



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk, urllib, re, gobject
from xl import common, xlmisc, library
from gettext import gettext as _
import collection

class EmptyDriver(object):
    def __init__(self):
        self.all = library.TrackData()

    def search_tracks(self, *e):
        return self.all

    def disconnect(self):
        pass

    def connect(self, *e):
        pass

class DeviceTransferQueue(gtk.VBox):
    """ 
        Shows tracks that are waiting to be transferred to the iPod
    """
    def __init__(self, panel):
        """
            Initializes the queue
        """
        gtk.VBox.__init__(self)
        self.panel = panel
        self.set_border_width(0)
        self.set_spacing(3)
        self.set_size_request(-1, 250)
        self.songs = []
        self.transferring = False
        self.stopped = True

        label = gtk.Label(_("Transfer Queue"))
        label.set_alignment(0, .50)
        self.pack_start(label, False, True)

        scroll = gtk.ScrolledWindow()
        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        view = gtk.TreeView()
        scroll.add(view)
        scroll.set_shadow_type(gtk.SHADOW_IN)
        self.list = xlmisc.ListBox(view)
        self.pack_start(scroll, True, True)

        self.progress = gtk.ProgressBar()
        self.pack_start(self.progress, False, False)

        buttons = gtk.HBox()
        buttons.set_spacing(3)
        self.clear = gtk.Button()
        image = gtk.Image()
        image.set_from_stock('gtk-clear', gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.clear.set_image(image)

        self.stop = gtk.Button()
        image = gtk.Image()
        image.set_from_stock('gtk-stop', gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.stop.set_image(image)
        self.stop.connect('clicked', self.on_stop)

        self.transfer = gtk.Button(_("Transfer"))
        buttons.pack_end(self.transfer, False, False)
        buttons.pack_end(self.clear, False, False)
        buttons.pack_end(self.stop, False, False)
        self.clear.connect('clicked', self.on_clear)
        self.transfer.connect('clicked', self.start_transfer)

        self.pack_start(buttons, False, False)
        targets = [('text/uri-list', 0, 0)]
        self.list.list.drag_dest_set(gtk.DEST_DEFAULT_ALL, targets,
            gtk.gdk.ACTION_COPY)
        self.list.list.connect('drag_data_received', self.drag_data_received)

    def check_transfer(self):
        """
            Checks to see if a transfer is in progress, and if so, it throws
            an error
        """
        if self.transferring:
            common.error(self.panel.exaile.window, _('A transfer is in '
                'progress, please wait for it to stop before attempting '
                'to perform this operation.'))
            return False

        return True

    def on_stop(self, *e):
        """
            Stops the transfer
        """
        self.transferring = False
        self.stopped = True

    def on_clear(self, widget):
        """
            Clears the queue
        """
        if not self.check_transfer(): return
        self.panel.queue = None
        self.hide()
        self.destroy()

    @common.threaded
    def start_transfer(self, widget):
        """
            Runs the transfer
        """
        if self.transferring: return
        self.transferring = True
        self.stopped = False
        gobject.idle_add(self.panel.exaile.status.set_first, _("Starting "
            "transfer..."), 3000)
        items = self.list.rows[:]
        total = len(self.list.rows)
        self.panel.transferring = True
        driver = self.panel.driver
        count = 0
        while True:
            if self.stopped: return
            if not items: break
            item = items.pop()
            driver.put_item(item)
            per = float(count) / float(total)
            count += 1
            gobject.idle_add(self.update_progress, item, per)
            xlmisc.log("set percent to %s" % per)

        if self.stopped: return
        gobject.idle_add(self.progress.set_fraction, 1)
        gobject.idle_add(self.panel.exaile.status.set_first, _("Finishing"
            " transfer..."), 3000)
        gobject.idle_add(self.panel.transfer_done)
        self.transferring = False

    def update_progress(self, song, percent):
        """
            Updates the progress of the transfer
        """
        self.list.remove(song)
        self.progress.set_fraction(percent)

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        """ 
            Called when a track is dropped in the transfer queue
        """
        # just pass it on to the iPodPanel
        if not self.check_transfer(): return
        self.panel.drag_data_received(tv, context, x, y, selection, info,
            etime)

class DeviceDragItem(object):
    def __init__(self, track, target):
        self.track = track
        self.target = target

    def __str__(self):
        return str(self.track)

class DevicePanel(collection.CollectionPanel):
    """
        Collection panel that allows for different collection drivers
    """
    name = 'device'
    def __init__(self, exaile):
        collection.CollectionPanel.__init__(self, exaile)
        self.driver = None
        self.drivers = {}
        self.all = library.TrackData()
        self.tree = xlmisc.DragTreeView(self, True, True)
        self.tree.set_headers_visible(False)

        self.chooser = self.xml.get_widget('device_driver_chooser')
        self.track_count = self.xml.get_widget('device_track_count')
        self.connect_button = self.xml.get_widget('device_connect_button')
        self.connect_button.connect('clicked', self.change_driver)

        self.store = gtk.ListStore(str, object)
        cell = gtk.CellRendererText()
        self.chooser.pack_start(cell)
        self.chooser.add_attribute(cell, 'text', 0)
        self.chooser.set_model(self.store)
        self.tree.set_row_separator_func(
                lambda m, i: m.get_value(i, 1) is None)

        container = self.xml.get_widget('%s_box' % self.name)

        scroll = gtk.ScrolledWindow()
        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scroll.add(self.tree)
        scroll.set_shadow_type(gtk.SHADOW_IN)
        container.pack_start(scroll, True, True)
        container.reorder_child(scroll, 3)
        container.show_all()

        selection = self.tree.get_selection()
        selection.set_mode(gtk.SELECTION_MULTIPLE)
        pb = gtk.CellRendererPixbuf()
        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Text')
        col.pack_start(pb, False)
        col.pack_start(cell, True)
        col.set_attributes(pb, pixbuf=0)
        self.tree.append_column(col)
        col.set_cell_data_func(cell, self.track_data_func)
        self.update_drivers(True)
        self.transferring = False
        self.connected = False
        self.queue = None
        self.chooser.set_active(0)

    def get_urls_for(self, items):
        driver_name = self.get_driver_name()
        return ["device_%s://%s" % (driver_name, urllib.quote(item.loc))
            for item in items]

    def done_loading_tree(self):
        """
            Called when the collection tree is done loading
        """
        if self.driver:
            if hasattr(self.driver, 'done_loading_tree'):
                self.driver.done_loading_tree()

    def show_device_panel_menu(self, widget, event, item):
        """
            Shows the device panel menu
        """
        if self.driver and hasattr(self.driver, 'get_menu'):
            menu = self.driver.get_menu(item, self.menu)
        else:
            menu = self.menu

        menu.popup(None, None, None, event.button, event.time)

    def remove_tracks(self, tracks):
        """ 
            Removes tracks from the current device
        """
        if not hasattr(self.driver, 'remove_tracks'):
            common.error(self.exaile.window, _("This device does "
                "not support removing tracks"))
            return

        self.driver.remove_tracks(tracks)

    def get_initial_root(self, model):
        if self.driver is not None and hasattr(self.driver, 
            'get_initial_root'):
            return getattr(self.driver, 'get_initial_root')(model)
        else:
            return None

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        self.tree.unset_rows_drag_dest()
        self.tree.drag_dest_set(gtk.DEST_DEFAULT_ALL, self.tree.targets, 
            gtk.gdk.ACTION_COPY)
        if not self.connected:
            common.error(self.exaile.window, _("Not connected to any media"
                " device"))
            return
        path = self.tree.get_path_at_pos(x, y)

        target = None
        if path:
            iter = self.model.get_iter(path[0])
            target = self.model.get_value(iter, 1)

        loc = selection.get_uris()
        items = []
        for url in loc:
            url = urllib.unquote(url)
            m = re.search(r'^device_(\w+)://', url)
            if m:
                song = self.get_song(url)
            else:
                song = library.read_track(self.exaile.db, self.exaile.all_songs, url)

            if song:
                items.append(DeviceDragItem(song, target))

        if items:
            self.add_to_transfer_queue(items)

    def add_to_transfer_queue(self, items):
        """
            Adds to the device transfer queue
        """
        if hasattr(self.driver, 'check_transfer_items'):
            if not self.driver.check_transfer_items(items):
                return
        if not hasattr(self.driver, 'put_item'):
            common.error(self.exaile.window, _("The current device "
                "does not support transferring music."))
            return
        if self.transferring:
            common.error(self.panel.exaile.window, _('A transfer is in '
                'progress, please wait for it to stop before attempting '
                'to perform this operation.'))
            return

        if not self.queue:
            self.queue_box = self.xml.get_widget('device_queue_box')
            self.queue = DeviceTransferQueue(self)
            self.queue_box.pack_start(self.queue, False, False)

        queue = self.queue.songs
        queue.extend(items)

        self.queue.list.set_rows(queue)
        if queue:
            self.queue.show_all()
        else:
            self.queue.hide()
            self.queue.destroy()
            self.queue = None

    def transfer_done(self):
        """
            called when the transfer is complete
        """
        if hasattr(self.driver, 'transfer_done'):
            self.driver.transfer_done()
        if self.queue:
            self.queue.hide()
            self.queue.destroy()
            self.queue = None
        self.transferring = None
        self.load_tree(True)

    def change_driver(self, button):
        """
            Changes the current driver
        """
        if self.driver and self.queue:
            if not self.queue.check_transfer():
                return
        if self.connected:
            self.driver.disconnect()
            self.driver = EmptyDriver()
            self.connected = False
            img = gtk.Image()
            img.set_from_stock('gtk-disconnect', gtk.ICON_SIZE_BUTTON)
            self.track_count.set_label(_("0 tracks"))
            self.load_tree(True)
            self.connect_button.set_image(img)
            return

        iter = self.chooser.get_active_iter()
        driver = self.store.get_value(iter, 1)
        if not isinstance(driver, EmptyDriver):
            self.connect(driver)
            img = gtk.Image()
            img.set_from_stock('gtk-connect', 
                gtk.ICON_SIZE_BUTTON)
            self.connect_button.set_image(img)

    def update_drivers(self, initial=False):
        """
            Updates the driver list
        """
        if not initial:
            self.exaile.show_device_panel(len(self.drivers) > 0)
        count = 1
        select = 0
        self.store.clear()
        # TRANSLATORS: No driver
        self.store.append([_('None'), EmptyDriver()])

        for k, v in self.drivers.iteritems():
            if k == self.driver:
                select = count

            self.store.append([v, k])            

        if select > 0:
            self.chooser.disconnect(self.change_id)
            self.chooser.set_active(select)
            self.change_id = self.chooser.connect('changed', self.change_driver)
        else:
            self.chooser.set_active(0)

    def add_driver(self, driver, name):
        if not self.drivers.has_key(driver):
            self.drivers[driver] = name
        self.update_drivers()

    def remove_driver(self, driver):
        if self.drivers.has_key(driver):
            del self.drivers[driver]
        self.update_drivers()

    def connect(self, driver):
        self.track_count.set_label(_("Connecting..."))
        try:
            driver.connect(self)
        except:
            common.error(self.exaile.window, _("Error connecting to device"))
            xlmisc.log_exception()
            self.on_connect_complete(None)

    def on_error(self, error):
        """
            Called when there is an error in a device driver during connect
        """
        common.error(self.exaile.window, error)
        self.on_connect_complete(None)
        
    def on_connect_complete(self, driver):
        """ 
            Called when the connection is complete
        """
        self.driver = driver
        if not self.driver:
            self.driver = EmptyDriver()
            self.connected = False
            img = gtk.Image()
            img.set_from_stock('gtk-disconnect', gtk.ICON_SIZE_BUTTON)
            self.connect_button.set_image(img)
        else:
            self.connected = True
        # TRANSLATORS: Number of tracks available on the device
        self.track_count.set_label(_("%d tracks") % len(self.driver.all))

        self.load_tree()

    def search_tracks(self, keyword, all=None):
        if not self.driver: self.all = library.TrackData()
        else: self.all = self.driver.search_tracks(keyword)
        return self.all

    def get_driver_name(self):
        if not self.driver: return None
        return self.driver.name

    def get_song(self, loc):
        song = self.all.for_path(loc.replace('device_%s://' % self.driver.name, ''))
        if hasattr(self.driver, 'get_song'):
            return self.driver.get_song(song)
        else:
           return song

    def load_tree(self, event=None):
        collection.CollectionPanel.load_tree(self, event)
        self.track_count.set_label(_("%d tracks") % len(self.driver.all))




#!/usr/bin/env python

import os, re, sys

h = open('debian/changelog')
line = h.readline()
h.close()

m = re.search('\(([^)]+)\)', line)
if not m:
    print "Could not determine version number, bailing!"
    sys.exit(1)

version = m.group(1)
print "Creating packages for exaile version: %s" % version

h = open('exaile.py')
lines = h.readlines()
h.close()

h = open('exaile.py', 'w')
for line in lines:
    line = re.sub("__version__ = '[^']+'", "__version__ = '%s'" % version,
        line)
    h.write(line)

h.close()

cur = os.getcwd()
branch = cur.split('/').pop()
os.chdir('../build')
os.system('cp -rf ../%s exaile_%s' % (branch, version))
os.chdir('exaile_%s' % version)

for m in ('.pyc', '.pyo', '.bzr'):
    os.system('find . -name %s -exec rm -rf {} \;' % m)

os.system('debuild -i -S -sa')
os.system('make clean')
os.chdir('..')

os.system('rm -rf exaile_%s' % version)

# create md5s
for item in ('.tar.gz', '_i386.deb'):
    os.system('md5sum exaile_%s%s > exaile_%s%s.md5' % (version, item,
        version, item))
os.chdir(cur)
print 'Packages have been created successfully'
